"""All sorts of form widgets, mainly those used in EditPanel.
"""
# from __future__ import annotations
# pylint: disable=maybe-no-member
import wx
from wx.lib.scrolledpanel import ScrolledPanel
from abc import abstractmethod
import copy
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, cast
from .config import get_theme, get_setting, Color
from .events import (DidModifyCompartmentsEvent, DidModifyNodesEvent, DidModifyReactionEvent,
                     DidMoveCompartmentsEvent, DidMoveNodesEvent, DidMoveReactionCenterEvent, DidResizeCompartmentsEvent,
                     DidResizeNodesEvent, post_event)
from .mvc import IController, ModifierTipStyle
from .utils import change_opacity, gchain, no_rzeros, on_msw, resource_path
from .canvas.canvas import Canvas, Node
from .canvas.data import ChoiceItem, Compartment, FONT_FAMILY_CHOICES, FONT_STYLE_CHOICES, FONT_WEIGHT_CHOICES, Reaction, TEXT_ALIGNMENT_CHOICES, LinePrim, PolygonPrim, Primitive, TEXT_POSITION_CHOICES, compute_centroid
from .canvas.geometry import Rect, Vec2, clamp_rect_pos, clamp_rect_size, get_bounding_rect, calc_node_dimensions
from .canvas.utils import get_nodes_by_idx, get_rxns_by_idx
from .canvas.data import CirclePrim, RectanglePrim, CompositeShape
from rkviewer.plugin import api


ColorCallback = Callable[[wx.Colour], None]
FloatCallback = Callable[[float], None]


def GetMultiEnum(entries: List[Any], fallback):
    """Similar to _GetMultiColor, but for enums.

    Need to specify a fallback value in case the entries are different.
    """
    entries_set = set(entries)
    if len(entries_set) == 1:
        return next(iter(entries_set))
    else:
        return fallback


def GetMultiFloatText(values: Set[float], precision: int) -> str:
    """Returns the common float value if the set has only one element, otherwise return "?".

    See _GetMultiColor for more detail.
    """
    return no_rzeros(next(iter(values)), precision) if len(values) == 1 else '?'


def GetMultiInt(values: Set[int]) -> Optional[int]:
    """Returns the common float value if the set has only one element, otherwise return "?".

    See _GetMultiColor for more detail.
    """
    return next(iter(values)) if len(values) == 1 else None


def GetMultiColor(colors: List[wx.Colour]) -> Tuple[wx.Colour, Optional[int]]:
    """Helper method for producing one single color from a list of colors.

    Editing programs that allows selection of multiple entities usually support editing all of
    the selected entities at once. When a property of all the selected entities are the same,
    the displayed value of that property is that single value precisely. However, if they are
    not the same, usually a "null" or default value is shown on the form. Following this scheme,
    this helper returns the common color/alpha if all values are the same, or a default value
    if not.

    Note:
        On Windows the RGB and the alpha are treated as different fields due to the lack of
        alpha field in the color picker screen. Therefore, the RGB and the alpha fields are
        considered different fields as far as uniqueness is considered.
    """
    if on_msw():
        rgbset = set(c.GetRGB() for c in colors)
        rgb = copy.copy(wx.Colour(127, 127, 127))
        if len(rgbset) == 1:
            rgb.SetRGB(next(iter(rgbset)))

        alphaset = set(c.Alpha() for c in colors)
        alpha = next(iter(alphaset)) if len(alphaset) == 1 else None
        return rgb, alpha
    else:
        rgbaset = set(c.GetRGBA() for c in colors)
        rgba = copy.copy(wx.Colour(127, 127, 127))
        if len(rgbaset) == 1:
            rgba.SetRGBA(next(iter(rgbaset)))

        return rgba, None


def AlphaToText(alpha: Optional[int], prec: int) -> str:
    """Simple helper for converting an alpha value ~[0, 255] to the range [0, 1].

    Args:
        alpha: The alpha value in range 0-255. If None, "?" will be returned.
        precision: The precision of the float string returned.
    """
    if alpha is None:
        return '?'
    else:
        return no_rzeros(alpha / 255, prec)


def _SetBestInsertion(ctrl: wx.TextCtrl, orig_text: str, orig_insertion: int):
    """Set the most natural insertion point for a paired-number text control.

    The format of the text control must be "X,Y" where X, Y are numbers, allowing whitespace.
    This should be called after the text control is autoly changed by View during user's
    editing. Normally if the text changes the caret will be reset to the 0th position, but this
    calculates a smarter position to place the caret to produce a more natural behavior.

    Args:
        ctrl: The text control, whose value is already programmatically changed.
        orig_text: The value of the text control before it was changed.
        orig_insertion: The original caret position from GetInsertionPoint().
    """
    new_text = ctrl.GetValue()
    try:
        mid = orig_text.index(',')
    except ValueError:
        # can't find comma; directly return
        return

    if orig_insertion > mid:
        ctrl.SetInsertionPoint(len(new_text))
    else:
        tokens = new_text.split(',')
        assert len(tokens) == 2

        left = tokens[0].strip()
        lstart = new_text.index(left)
        lend = lstart + len(left)
        ctrl.SetInsertionPoint(lend)


def ChangePairValue(ctrl: wx.TextCtrl, new_val: Vec2, prec: int):
    """Helper for updating the value of a paired number TextCtrl.

    The TextCtrl accepts text in the format "X, Y" where X and Y are floats. The control is
    not updated if the new and old values are identical (considering precision).

    Args:
        ctrl: The TextCtrl widget.
        new_val: The new pair of floats to update the control with.
        prec: The precision of the numbers. The new value is rounded to this precision.
    """
    old_text = ctrl.GetValue()
    old_val = Vec2(parse_num_pair(old_text))

    # round old_val to desired precision. We don't want to refresh value when user is typing,
    # even if their value exceeded our precision
    if old_val != new_val:
        if ctrl.HasFocus():
            orig_insertion = ctrl.GetInsertionPoint()
            wx.CallAfter(
                lambda: _SetBestInsertion(ctrl, old_text, orig_insertion))
        ctrl.ChangeValue('{} , {}'.format(
            no_rzeros(new_val.x, prec), no_rzeros(new_val.y, prec)))


def parse_num_pair(text: str) -> Optional[Tuple[float, float]]:
    """Parse a pair of floats from a string with form "X,Y" and return a tuple.

    Returns None if failed to parse.
    """
    nums = text.split(",")
    if len(nums) != 2:
        return None

    xstr, ystr = nums
    x = None
    y = None
    try:
        x = float(xstr)
        y = float(ystr)
    except ValueError:
        return None

    return (x, y)


def parse_precisions(text: str) -> Tuple[int, int]:
    """Given a string in format 'X, Y' of floats, return the decimal precisions of X and Y."""
    nums = text.split(",")
    assert len(nums) == 2

    xstr = nums[0].strip()
    ystr = nums[1].strip()
    x_prec = None
    try:
        x_prec = len(xstr) - xstr.index('.') - 1
    except ValueError:
        x_prec = 0

    y_prec = None
    try:
        y_prec = len(xstr) - ystr.index('.') - 1
    except ValueError:
        y_prec = 0

    return (x_prec, y_prec)


class FieldGrid(wx.Window):
    def __init__(self, parent, form: 'EditPanelForm'):
        super().__init__(parent)
        self.SetForegroundColour(get_theme('toolbar_fg'))
        self.SetBackgroundColour(get_theme('toolbar_bg'))
        self.form = form
        self.labels = dict()
        self.badges = dict()
        self._label_font = wx.Font(wx.FontInfo().Bold())
        info_image = wx.Image(resource_path('info-2-16.png'), wx.BITMAP_TYPE_PNG)
        self._info_bitmap = wx.Bitmap(info_image)
        self._info_length = 16
        sizer = self.InitAndGetSizer(self.GetParent().GetParent().GetSize()[0])
        self.SetSizer(sizer)

    def InitAndGetSizer(self, edit_panel_width) -> wx.GridSizer:
        VGAP = 8
        HGAP = 5
        MORE_LEFT_PADDING = 0  # Left padding in addition to vgap
        MORE_TOP_PADDING = 2  # Top padding in addition to hgap
        MORE_RIGHT_PADDING = 0

        sizer = wx.GridBagSizer(vgap=VGAP, hgap=HGAP)

        # Set paddings
        # Add spacer of width w on the 0th column; add spacer of height h on the 0th row.
        # This results in a left padding of w + hgap and a top padding of h + vgap
        sizer.Add(MORE_LEFT_PADDING, MORE_TOP_PADDING, wx.GBPosition(0, 0), wx.GBSpan(1, 1))
        # Add spacer on column 3 to reserve space for info badge
        sizer.Add(self._info_length, 0, wx.GBPosition(0, 3), wx.GBSpan(1, 1))
        # Add spacer of width 5 on the 3rd column. This results in a right padding of 5 + hgap
        sizer.Add(MORE_RIGHT_PADDING, 0, wx.GBPosition(0, 4), wx.GBSpan(1, 1))

        # Ensure the input field takes up some percentage of width
        # Note that we might want to adjust this when scrollbars are displayed, but only in case
        # there is not enough width to display everything
        right_width = (edit_panel_width - VGAP * 3 - MORE_LEFT_PADDING - MORE_RIGHT_PADDING -
                       self._info_length) * .7
        sizer.Add(int(right_width), 0, wx.GBPosition(0, 2), wx.GBSpan(1, 1))
        sizer.AddGrowableCol(0, 3)
        sizer.AddGrowableCol(1, 7)
        return sizer

    def AppendControl(self, label_str: str, ctrl: wx.Control):
        """Append a control, its label, and its info badge to the last row of the sizer.

        Returns the automaticaly created label and info badge (wx.StaticText for now).
        """
        sizer = self.GetSizer()

        label = wx.StaticText(self, label=label_str)
        label.SetFont(self._label_font)
        rows = sizer.GetRows()
        sizer.Add(label, wx.GBPosition(rows, 1), wx.GBSpan(1, 1),
                  flag=wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT)
        sizer.Add(ctrl, wx.GBPosition(rows, 2), wx.GBSpan(1, 1),
                  flag=wx.ALIGN_CENTER_VERTICAL | wx.EXPAND)
        sizer.Add(0, self._info_length, wx.GBPosition(rows, 4), wx.GBSpan(1, 1))

        info_badge = wx.StaticBitmap(self, bitmap=self._info_bitmap)
        info_badge.Show(False)
        sizer.Add(info_badge, wx.GBPosition(rows, 3), wx.GBSpan(1, 1), flag=wx.ALIGN_CENTER)
        self.labels[ctrl.GetId()] = label
        self.badges[ctrl.GetId()] = info_badge

    def AppendSpacer(self, height: int, sizer=None):
        """Append a horizontal spacer with the given height.

        Note:
            The VGAP value still applies, i.e. there is an additional gap between the spacer and
            the next row.
        """
        if sizer is None:
            sizer = self.GetSizer()
        rows = sizer.GetRows()
        sizer.Add(0, height, wx.GBPosition(rows, 0), wx.GBSpan(1, 5))

    def AppendLine(self):
        """Append a horizontal spacer with the given height.

        Note:
            The VGAP value still applies, i.e. there is an additional gap between the spacer and
            the next row.
        """
        sizer = self.GetSizer()
        rows = sizer.GetRows()
        line = wx.StaticLine(self)
        sizer.Add(line, wx.GBPosition(rows, 0), wx.GBSpan(1, 5))

    def AppendSubtitle(self, text: str, add_spacers: bool = True) -> wx.StaticText:
        sizer = self.GetSizer()
        if add_spacers:
            self.AppendSpacer(3)
        sizer.Add(0, 0, wx.GBPosition(sizer.GetRows(), 0))
        statictext = wx.StaticText(self, label=text)
        font = wx.Font(wx.FontInfo(9))
        statictext.SetFont(font)
        sizer.Add(statictext, wx.GBPosition(sizer.GetRows(), 0),
                  wx.GBSpan(1, 5), flag=wx.ALIGN_CENTER)
        if add_spacers:
            self.AppendSpacer(0)
        return statictext

    def SetValidationState(self, good: bool, ctrl_id: str, message: str = ""):
        """Set the validation state for a control.

        Args:
            good: Whether the control is currently valid.
            ctrl_id: The ID of the control.
            message: The message displayed, if the control is not valid.
        """
        # self.Freeze()
        badge = self.badges[ctrl_id]
        if good:
            badge.Show(False)
        else:
            badge.Show(True)
            badge.SetToolTip(message)
        self.Layout()
        # self.Thaw()

    def CreateTextCtrl(self, **kwargs):
        """Create a text control that confirms to the theme."""
        if get_theme('text_field_border'):
            style = 0
        else:
            style = wx.BORDER_NONE
        ctrl = wx.TextCtrl(self, style=style, **kwargs)
        ctrl.SetBackgroundColour(get_theme('text_field_bg'))
        ctrl.SetForegroundColour(get_theme('text_field_fg'))
        return ctrl

    def CreateSpinCtrl(self, **kwargs):
        """Create a text control that confirms to the theme."""
        if get_theme('text_field_border'):
            style = 0
        else:
            style = wx.BORDER_NONE
        ctrl = wx.SpinCtrl(self, style=style, **kwargs)
        ctrl.SetBackgroundColour(get_theme('text_field_bg'))
        ctrl.SetForegroundColour(get_theme('text_field_fg'))
        return ctrl

    def CreateColorControl(self, label: str, alpha_label: str,
                           color_callback: ColorCallback, alpha_callback: FloatCallback,
                           alpha_range: Tuple[float, float] = (0, 1),
                           placeholder: wx.Colour = wx.Colour(127, 127, 127), placeholder_alpha=None) \
            -> Tuple[wx.ColourPickerCtrl, Optional[wx.TextCtrl]]:
        """Helper method for creating a color control and adding it to the form.

        Args:
            label: The label text for the color control.
            alpha_label: The label text for the alpha control. Relevant only on Windows.
            color_callback: Callback called when the color changes.
            alpha_callback: Callback called when the alpha changes. Relevant only on Windows.
            sizer: The sizer to which widgets should be added.
            alpha_range: The inclusive range for the alpha value.

        Returns:
            A tuple of the color control and the alpha control.
        """
        # Update placeholder to include alpha
        if placeholder_alpha:
            placeholder = wx.Colour(placeholder.Red(), placeholder.Green(), placeholder.Blue(),
                                    placeholder_alpha)

        ctrl = wx.ColourPickerCtrl(self)
        ctrl.SetColour(placeholder)
        ctrl.Bind(wx.EVT_COLOURPICKER_CHANGED, lambda e: color_callback(e.GetColour()))
        self.AppendControl(label, ctrl)

        alpha_ctrl = None

        if on_msw():
            # Windows does not support picking alpha in color picker. So we add an additional
            # field for that
            alpha_text = AlphaToText(placeholder_alpha, 2)
            alpha_ctrl = self.CreateTextCtrl(value=alpha_text)
            self.AppendControl(alpha_label, alpha_ctrl)
            callback = self.MakeFloatCtrlFunction(alpha_ctrl.GetId(), alpha_callback, alpha_range)
            alpha_ctrl.Bind(wx.EVT_TEXT, callback)

        return ctrl, alpha_ctrl

    def MakeFloatCtrlFunction(self, ctrl_id: str, callback: FloatCallback,
                              range_: Tuple[Optional[float], Optional[float]],
                              left_incl: bool = True, right_incl: bool = True):
        """Helper method that creates a validation function for a TextCtrl that only allows floats.

        Args:
            ctrl_id: ID of the TextCtrl, for which this validation function is created.
            callback: Callback for when the float is changed and passes the validation tests.
            range_: Inclusive range for the allowed floats.

        Returns:
            The validation function.
        """
        lo, hi = range_

        def float_ctrl_fn(evt):
            text = evt.GetString()
            value: float
            try:
                value = float(text)
            except ValueError:
                self.SetValidationState(False, ctrl_id, "Value must be a number")
                return

            good = True
            if left_incl:
                if lo is not None and value < lo:
                    good = False
            else:
                if lo is not None and value <= lo:
                    good = False

            if right_incl:
                if hi is not None and value > hi:
                    good = False
            else:
                if hi is not None and value >= hi:
                    good = False

            if not good:
                err_msg: str
                if lo is not None and hi is not None:
                    left = '[' if left_incl else '('
                    right = ']' if right_incl else ')'
                    err_msg = "Value must be in range {}{}, {}{}".format(left, lo, hi, right)
                else:
                    if lo is not None:
                        incl_text = 'or equal to ' if left_incl else ''
                        err_msg = "Value must greater than {}{}".format(incl_text, lo)
                    else:
                        incl_text = 'or equal to' if right_incl else ''
                        err_msg = "Value must less than {} {}".format(incl_text, hi)
                self.SetValidationState(False, ctrl_id, err_msg)
                return

            callback(value)
            self.SetValidationState(True, ctrl_id)

        return float_ctrl_fn

class PrimitiveGrid(FieldGrid):
    form: 'NodeForm'
    def __init__(self, parent, form: 'NodeForm'):
        super().__init__(parent, form)
        self.update_callbacks = list()
        # need another GetParent() in addition to FieldGrid's call
        sizer = self.InitAndGetSizer(self.GetParent().GetParent().GetParent().GetSize()[0])
        self.SetSizer(sizer)

    def UpdateValues(self, nodes):
        '''Update the values in the primitive fields.

        Requires:
            The FieldGrid contains the up-to-date field widgets for the given composite shape.
        '''
        for callback in self.update_callbacks:
            callback(nodes)


    def ColorPrimitiveControl(self, label: str, alpha_label: str, prop_name: str,
                              prim_index: int):
        '''Create a control for a color property.

        If prim_index is -1, then update the text primitive instead.
        '''

        def color_callback(value: wx.Colour):
            node_indices = self.form.selected_idx
            nodes = self.form.selected_nodes
            prims = self._GetPrimitives(nodes, prim_index)
            old_colors = [getattr(p, prop_name).to_wxcolour() for p in prims]

            self.form.self_changes = True
            with self.form.controller.group_action():
                for i, nodei in enumerate(node_indices):
                    # only update the RGB, not alpha
                    old_color = old_colors[i]
                    new_color = Color(value.Red(), value.Green(), value.Blue(), old_color.Alpha())
                    self.form.controller.set_node_primitive_property(self.form.net_index, nodei, prim_index,
                                                                    prop_name, new_color)

        def alpha_callback(value: float):
            node_indices = self.form.selected_idx
            prims = self._GetPrimitives(self.form.selected_nodes, prim_index)
            old_colors = [getattr(p, prop_name).to_wxcolour() for p in prims]

            self.form.self_changes = True
            with self.form.controller.group_action():
                for i, nodei in enumerate(node_indices):
                    old_color = old_colors[i]
                    new_color = Color(old_color.Red(), old_color.Green(),
                                    old_color.Blue(), int(255 * value))
                    self.form.controller.set_node_primitive_property(self.form.net_index, nodei, prim_index,
                                                                    prop_name, new_color)

        ctrl, alpha_ctrl = self.CreateColorControl(label, alpha_label, color_callback, alpha_callback)

        # callback for when the canvs is upated by user input
        def update_cb(nodes: List[Node]):
            prims = self._GetPrimitives(nodes, prim_index)
            old_colors = [getattr(p, prop_name).to_wxcolour() for p in prims]
            color_union, alpha_union = GetMultiColor(old_colors)
            self.form.self_changes = True
            ctrl.SetColour(color_union)
            if alpha_ctrl:
                alpha_ctrl.ChangeValue(AlphaToText(alpha_union, 2))

        self.update_callbacks.append(update_cb)


    def FloatPrimitiveControl(self, label: str, prop_name: str, prim_index: int):
        '''Create a control for a floating point property.

        If prim_index is -1, then update the text primitive instead.
        '''

        def callback(value: float):
            node_indices = self.form.selected_idx

            self.form.self_changes = True
            with self.form.controller.group_action():
                for nodei in node_indices:
                    # only update the RGB, not alpha
                    self.form.controller.set_node_primitive_property(self.form.net_index, nodei, prim_index,
                                                                    prop_name, value)

        # TODO update values not here
        text_ctrl = self.CreateTextCtrl()
        outer_callback = self.MakeFloatCtrlFunction(text_ctrl.GetId(),
                                                    callback, (0, None), left_incl=False)
        text_ctrl.Bind(wx.EVT_TEXT, outer_callback)
        self.AppendControl(label, text_ctrl)

        def update_cb(nodes: List[Node]):
            prims = self._GetPrimitives(nodes, prim_index)
            old_values = [getattr(p, prop_name) for p in prims]
            update_value = GetMultiFloatText(set(old_values), 2)
            text_ctrl.ChangeValue(update_value)

        self.update_callbacks.append(update_cb)

    def IntPrimitiveControl(self, label: str, prop_name: str,
                            prim_index: int, min_=0, max_=100):
        '''Create a control for a floating point property.

        If prim_index is -1, then update the text primitive instead.
        '''
        def spin_callback(value: int):
            node_indices = self.form.selected_idx
            self.form.self_changes = True
            with self.form.controller.group_action():
                for nodei in node_indices:
                    # only update the RGB, not alpha
                    self.form.controller.set_node_primitive_property(self.form.net_index, nodei, prim_index,
                                                                    prop_name, value)

        def text_callback(value: str):
            if value:
                spin_callback(int(value))


        int_ctrl = self.CreateSpinCtrl(min=min_, max=max_)
        int_ctrl.Bind(wx.EVT_SPINCTRL, lambda e: spin_callback(e.GetInt()))
        int_ctrl.Bind(wx.EVT_TEXT, lambda e: text_callback(e.GetString()))

        self.AppendControl(label, int_ctrl)

        def update_cb(nodes: List[Node]):
            prims = self._GetPrimitives(self.form.selected_nodes, prim_index)
            old_values = [getattr(p, prop_name) for p in prims]
            updated_value = GetMultiInt(set(old_values)) or 0
            int_ctrl.SetValue(updated_value)

        self.update_callbacks.append(update_cb)

    def ChoicePrimitiveControl(self, label: str, prop_name: str, prim_index: int,
                               choice_items: List[ChoiceItem]):
        # TODO set original value
        def callback(e):
            node_indices = self.form.selected_idx
            index = e.GetInt()
            value = choice_items[index].value
            self.form.self_changes = True
            with self.form.controller.group_action():
                for nodei in node_indices:
                    self.form.controller.set_node_primitive_property(self.form.net_index, nodei, prim_index,
                                                                    prop_name, value)
        texts = [item.text for item in choice_items]
        choice_ctrl = wx.Choice(self, choices=texts)

        choice_ctrl.Bind(wx.EVT_CHOICE, callback)
        self.AppendControl(label, choice_ctrl)

        def update_cb(nodes):
            prims = self._GetPrimitives(nodes, prim_index)
            old_values = set(getattr(prim, prop_name) for prim in prims)

            # Set commonly selected item
            if len(old_values) == 1:
                # Find choice item with given value
                sel_ind = None
                old_value = next(iter(old_values))
                for index, item in enumerate(choice_items):
                    if item.value == old_value:
                        sel_ind = index
                        break
                else:
                    assert False, "This should never happen"
                choice_ctrl.SetSelection(sel_ind)
            else:
                # Different values; set dropdown to none
                choice_ctrl.SetSelection(wx.NOT_FOUND)
        self.update_callbacks.append(update_cb)

    def _GetPrimitives(self, nodes, prim_index):
        if prim_index == -1:
            return [n.composite_shape.text_item[0] for n in nodes]
        return [n.composite_shape.items[prim_index][0] for n in nodes]


class PrimitiveSection(wx.Window):
    subsections: List[PrimitiveGrid]

    def __init__(self, node_form: 'NodeForm', com_shape: CompositeShape):
        super().__init__(node_form)
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizerflags = wx.SizerFlags().Expand()
        self.update_callbacks = list()
        self.form = node_form

        self.subsections = list()
        # self._primitives_heading = self.main_section.AppendSubtitle('Shape properties')
        # self.main_section.AppendSpacer(0)
        # node_indices = [n.index for n in nodes]
        for prim_index in range(len(com_shape.items)):
            # primitives = [cs.items[prim_index][0] for cs in com_shapes]
            one_prim = com_shape.items[prim_index][0]
            subtitle_text = '{name} ({idx})'.format(idx=prim_index + 1, name=one_prim.name)
            subsection = PrimitiveGrid(self, node_form)
            self.subsections.append(subsection)
            subsection.AppendSubtitle(subtitle_text)
            if isinstance(one_prim, RectanglePrim):
                subsection.ColorPrimitiveControl('fill color', 'fill opacity', 'fill_color',
                                                 prim_index)
                subsection.ColorPrimitiveControl('border color', 'border opacity', 'border_color',
                                                 prim_index)
                subsection.FloatPrimitiveControl('border width', 'border_width',
                                                 prim_index)
                subsection.FloatPrimitiveControl('corner radius', 'corner_radius',
                                                 prim_index)
            elif isinstance(one_prim, LinePrim):
                subsection.ColorPrimitiveControl('line color', 'line opacity', 'border_color',
                                                 prim_index)
                subsection.FloatPrimitiveControl('line width', 'border_width',
                                                 prim_index)
            elif isinstance(one_prim, CirclePrim) or isinstance(one_prim, PolygonPrim):
                subsection.ColorPrimitiveControl('fill color', 'fill opacity', 'fill_color',
                                                 prim_index)
                subsection.ColorPrimitiveControl('border color', 'border opacity', 'border_color',
                                                 prim_index)
                subsection.FloatPrimitiveControl('border width', 'border_width',
                                                 prim_index)

        subtitle_text = 'Text'
        subsection = PrimitiveGrid(self, node_form)
        self.subsections.append(subsection)
        subsection.AppendSubtitle(subtitle_text)
        # Create text primitive
        subsection.IntPrimitiveControl('font size', 'font_size', -1, min_=1, max_=100)
        subsection.ColorPrimitiveControl('font color', 'font opacity', 'font_color',
                                         -1)
        subsection.ColorPrimitiveControl('highlight color', 'highlight opacity', 'bg_color',
                                         -1)
        subsection.ChoicePrimitiveControl('font family', 'font_family', -1, FONT_FAMILY_CHOICES)
        subsection.ChoicePrimitiveControl('font style', 'font_style', -1, FONT_STYLE_CHOICES)
        subsection.ChoicePrimitiveControl('font weight', 'font_weight', -1, FONT_WEIGHT_CHOICES)
        subsection.ChoicePrimitiveControl('alignment', 'alignment', -1, TEXT_ALIGNMENT_CHOICES)
        subsection.ChoicePrimitiveControl('text position', 'position', -1, TEXT_POSITION_CHOICES)

        for subsection in self.subsections:
            sizer.Add(subsection, sizerflags)
        self.SetSizer(sizer)

    def UpdatePrimitiveValues(self):
        selected_nodes = self.form.selected_nodes
        for subsection in self.subsections:
            subsection.UpdateValues(selected_nodes)


class EditPanelForm(ScrolledPanel):
    """Base class for a form to be displayed on the edit panel.

    Attributes:
        ColorCallback: Callback type for when a color input is changed.
        FloatCallback: Callback type for when a float input is changed.
        canvas: The associated canvas.
        controller: The associated controller.
        net_index: The current network index. For now it is 0 since there is only one tab.
    """

    canvas: Canvas
    controller: IController
    net_index: int
    labels: Dict[str, wx.Window]
    badges: Dict[str, wx.Window]
    sections: List[FieldGrid]
    _label_font: wx.Font  #: font for the form input label.
    _info_bitmap: wx.Bitmap  # :  bitmap for the info badge (icon), for when an input is invalid.
    _info_length: int  #: length of the square reserved for _info_bitmap
    _title: wx.StaticText  #: title of the form
    self_changes: bool  #: flag for if edits were made but the controller hasn't updated the view yet

    def __init__(self, parent, canvas: Canvas, controller: IController):
        #super().__init__(parent, style=wx.VSCROLL)
        super().__init__(parent, style = wx.ALWAYS_SHOW_SB)
        self.SetForegroundColour(get_theme('toolbar_fg'))
        self.SetBackgroundColour(get_theme('toolbar_bg'))
        self.canvas = canvas
        self.controller = controller
        self.net_index = 0
        self._title = wx.StaticText(self, style=wx.ALIGN_CENTER)  # only displayed when node(s) are selected
        title_font = wx.Font(wx.FontInfo(10))
        self._title.SetFont(title_font)
        self.self_changes = False
        self._selected_idx = set()

    # OVerride the ScrolledPanel behavior of jumping to the child that has the focus
    def OnChildFocus(self, evt):
        pass

    @property
    def selected_idx(self):
        return self._selected_idx

    def CreateChildren(self):
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizerflags = wx.SizerFlags().Expand()

        sizer.Add(self._title, sizerflags.Border(wx.BOTTOM, 5))
        sizer.Add(wx.StaticLine(self), sizerflags)

        for section in self.sections:
            sizer.Add(section, sizerflags)
        self.CreateControls()

        self.SetSizer(sizer)
        self.SetupScrolling()

    @abstractmethod
    def UpdateAllFields(self):
        pass

    @abstractmethod
    def CreateControls(self):
        pass

    def ExternalUpdate(self):
        if len(self._selected_idx) != 0 and not self.self_changes:
            self.UpdateAllFields()

        # clear validation errors
        # TODO
        for section in self.sections:
            for id in section.badges.keys():
                section.SetValidationState(True, id)
        self.self_changes = False


class NodeForm(EditPanelForm):
    """Form for editing one or multiple nodes.

    Attributes:
    """
    #_compartments: List[Compartment]
    contiguous: bool
    id_ctrl: wx.TextCtrl
    conc_ctrl: wx.TextCtrl
    pos_ctrl: wx.TextCtrl
    size_ctrl: wx.TextCtrl
    nodeStatusDropDown: wx.Choice
    compositeShapesDropDown: wx.Choice
    lockNodeCheckBox: wx.CheckBox

    _nodes: List[Node]  #: current list of nodes in canvas.
    _selected_idx: Set[int]  #: current list of selected indices in canvas.
    _bounding_rect: Optional[Rect]  #: the exact bounding rectangle of the selected nodes

    def __init__(self, parent, canvas: Canvas, controller: IController):
        super().__init__(parent, canvas, controller)
        self.all_nodes = list()
        #self.compartments = list()
        self.main_section = FieldGrid(self, self)
        self.sections = [self.main_section]
        self._bounding_rect = None  # No padding
        # boolean to indicate whether only nodes are selected, and only nodes from the same
        # compartment are selected
        self.contiguous = True
        self.last_prim_section = None
        self.prim_section_cache = dict()

        self.CreateChildren()

    @property
    def selected_nodes(self):
        return [n for n in self.all_nodes if n.index in self.selected_idx]

    def UpdateNodes(self, nodes: List[Node]):
        """Function called after the list of nodes have been updated."""
        self.all_nodes = nodes
        self._UpdateBoundingRect()
        self.ExternalUpdate()

    def NodesMovedOrResized(self, evt):
        """Called when nodes are moved or resized by dragging"""
        if not evt.dragged:
            return
        # Possibly no nodes are selected because they are moved along with the compartments
        if len(self._selected_idx) != 0:
            self._UpdateBoundingRect()
            prec = 2
            ChangePairValue(self.pos_ctrl, self._bounding_rect.position, prec)
            ChangePairValue(self.size_ctrl, self._bounding_rect.size, prec)

    def _UpdateBoundingRect(self):
        """Update bounding rectangle; mixed indicates whether both nodes and comps are selected.
        """
        rects = [n.rect for n in self.all_nodes if n.index in self._selected_idx]
        # It could be that compartments have been updated but selected indices have not.
        # In that case rects can be empty
        if len(rects) != 0:
            self._bounding_rect = get_bounding_rect(rects)

    
    # def UpdateCompartments(self, comps: List[Compartment]):
    #     self.compartments = comps
    #     self._UpdateBoundingRect()
    #     self.ExternalUpdate()

    def UpdateSelection(self, selected_idx: Set[int], comps_selected: bool):
        """Function called after the list of selected nodes have been updated."""
        self._selected_idx = selected_idx
        if len(selected_idx) != 0:
            self._UpdateBoundingRect()

        if comps_selected:
            self.contiguous = False
        else:
            nodes = self.selected_nodes
            self.contiguous = len(set(n.comp_idx for n in nodes)) <= 1

        if len(self._selected_idx) != 0:
            # clear position value
            self.pos_ctrl.ChangeValue('')
            self.UpdateAllFields()

            title_label = 'Edit Node' if len(self._selected_idx) == 1 else 'Edit Multiple Nodes'
            self._title.SetLabel(title_label)

            id_text = 'identifier' if len(self._selected_idx) == 1 else 'identifiers'
            self.main_section.labels[self.id_ctrl.GetId()].SetLabel(id_text)

            comp_text = 'compartment' if len(self._selected_idx) == 1 else 'compartment'
            self.main_section.labels[self.comp_ctrl.GetId()].SetLabel(comp_text)

            name_text = 'name' if len(self._selected_idx) == 1 else 'name'
            self.main_section.labels[self.name_ctrl.GetId()].SetLabel(name_text)

            SBO_text = 'SBO' if len(self._selected_idx) == 1 else 'SBO'
            self.main_section.labels[self.SBO_ctrl.GetId()].SetLabel(SBO_text)

            concentration_text = 'concentration' if len(self._selected_idx) == 1 else 'concentrations'
            self.main_section.labels[self.conc_ctrl.GetId()].SetLabel(concentration_text)

            size_text = 'size' if len(self._selected_idx) == 1 else 'size'
            self.main_section.labels[self.size_ctrl.GetId()].SetLabel(size_text)

        self.ExternalUpdate()

    def CreateControls(self):
        self.id_ctrl = self.main_section.CreateTextCtrl()
        self.id_ctrl.Bind(wx.EVT_TEXT, self._OnIdText)
        self.main_section.AppendControl('identifier', self.id_ctrl)

        self.comp_ctrl = self.main_section.CreateTextCtrl()
        #self.comp_ctrl.Bind(wx.EVT_TEXT, self._OnCompText)
        self.main_section.AppendControl('compartment', self.comp_ctrl)

        self.name_ctrl = self.main_section.CreateTextCtrl()
        self.name_ctrl.Bind(wx.EVT_TEXT, self._OnNameText)
        self.main_section.AppendControl('name', self.name_ctrl)

        self.SBO_ctrl = self.main_section.CreateTextCtrl()
        self.SBO_ctrl.Bind(wx.EVT_TEXT, self._OnSBOText)
        self.main_section.AppendControl('SBO', self.SBO_ctrl)
        
        self.conc_ctrl = self.main_section.CreateTextCtrl()
        self.conc_ctrl.Bind(wx.EVT_TEXT, self._OnConcText)
        self.main_section.AppendControl('concentration', self.conc_ctrl)

        self.pos_ctrl = self.main_section.CreateTextCtrl()
        self.pos_ctrl.Bind(wx.EVT_TEXT, self._OnPosText)
        self.main_section.AppendControl('position', self.pos_ctrl)

        self.size_ctrl = self.main_section.CreateTextCtrl()
        #self.size_ctrl.Bind(wx.EVT_TEXT, self._OnSizeText)
        self.size_ctrl.Bind(wx.EVT_TEXT, self._OnNodeSizeText)
        self.main_section.AppendControl('size', self.size_ctrl)

        self.nodeStates = ['Floating Node', 'Boundary Node']
        self.nodeStatusDropDown = wx.Choice(self.main_section, choices=self.nodeStates)
        self.main_section.AppendControl('node status', self.nodeStatusDropDown)
        self.nodeStatusDropDown.Bind(wx.EVT_CHOICE, self.OnNodeStatusChoice)

        self.lockNodeCheckBox = wx.CheckBox(self.main_section, label='')
        self.main_section.AppendControl('lock node', self.lockNodeCheckBox)
        self.lockNodeCheckBox.Bind(wx.EVT_CHECKBOX, self.OnNodeLockCheckBox)

        self.compShapeNames = [
            x.name for x in self.controller.get_composite_shape_list(self.net_index)]
        self.compositeShapesDropDown = wx.Choice(self.main_section, choices=self.compShapeNames)
        self.main_section.AppendControl('shape', self.compositeShapesDropDown)
        self.compositeShapesDropDown.Bind(wx.EVT_CHOICE, self.OnCompositeShapes)

    def _OnIdText(self, evt):
        """Callback for the ID control."""
        new_id = evt.GetString()
        assert len(self._selected_idx) == 1
        [nodei] = self._selected_idx
        ctrl_id = self.id_ctrl.GetId()
        if len(new_id) == 0:
            self.main_section.SetValidationState(False, ctrl_id, "ID cannot be empty")
            return
        else:
            for node in self.all_nodes:
                if node.id == new_id:
                    self.main_section.SetValidationState(False, ctrl_id, "Not saved: Duplicate ID")
                    return
            else:
                # loop terminated fine. There is no duplicate ID
                self.self_changes = True
                with self.controller.group_action():
                    self.controller.rename_node(self.net_index, nodei, new_id)
                    post_event(DidModifyNodesEvent([nodei]))
        self.main_section.SetValidationState(True, self.id_ctrl.GetId())

    def _OnNameText(self, evt):
        """Callback for the name control."""
        new_name = evt.GetString()
        assert len(self._selected_idx) == 1
        [nodei] = self._selected_idx
        ctrl_name = self.name_ctrl.GetId()

        if len(new_name) != 0:
            self.self_changes = True
            with self.controller.group_action():
                self.controller.set_node_name(self.net_index, nodei, new_name)
                post_event(DidModifyNodesEvent([nodei]))
        self.main_section.SetValidationState(True, self.id_ctrl.GetId())

    def _OnSBOText(self, evt):
        """Callback for the SBO control."""
        new_SBO = evt.GetString()
        assert len(self._selected_idx) == 1
        [nodei] = self._selected_idx
        ctrl_SBO = self.SBO_ctrl.GetId()

        if len(new_SBO) != 0:
            if new_SBO[:8] == "SBO:0000" and new_SBO[8:11].isdigit():
                self.self_changes = True
                with self.controller.group_action():
                    self.controller.set_node_SBO(self.net_index, nodei, new_SBO)
                    post_event(DidModifyNodesEvent([nodei]))
            else:
                self.main_section.SetValidationState(False, ctrl_SBO, "Not saved: Invalid SBO Term ID (Example: SBO:0000247)")
                return
         
        self.main_section.SetValidationState(True, self.SBO_ctrl.GetId())

    # def _OnCompText(self, evt):
    #     """Callback for the compartment control."""
    #     #needs to double check especially comp_idx vs comp_id
    #     new_comp = evt.GetString()
    #     assert len(self._selected_idx) == 1
    #     [nodei] = self._selected_idx
    #     ctrl_comp = self.comp_ctrl.GetId()

    #     flag_comp_exist = 0
    #     allCompartments = api.get_compartments(self.net_index)
    #     for comp in allCompartments:
    #         if comp.index == new_comp:
    #             flag_comp_exist = 1
    #     if flag_comp_exist == 0:
    #         self.main_section.SetValidationState(False, ctrl_comp, "Not saved: Compartment ID does not exist.")
    #         return

    #     self.self_changes = True
    #     with self.controller.group_action():
    #         self.controller.set_compartment_of_node(self.net_index, nodei, new_comp)
    #         post_event(DidModifyNodesEvent([nodei]))
    #     self.main_section.SetValidationState(True, self.comp_ctrl.GetId())

    def _OnConcText(self, evt):
        """ Callback for the concentration control. """
        new_conc = evt.GetString()
        assert len(self._selected_idx) == 1 # TODO should work for multiple nodes
        [nodei] = self._selected_idx
        ctrl_conc = self.conc_ctrl.GetId()
        if len(new_conc) == 0:
            self.main_section.SetValidationState(False, ctrl_conc, "Concentration cannot be empty")
            return
        try:
            new_conc_float = float(new_conc)
        except ValueError:
            self.main_section.SetValidationState(False, ctrl_conc, "Concentration must be a numerical value") # TODO good message?
            return
        if new_conc_float < 0.0:
            self.main_section.SetValidationState(False, ctrl_conc, "Concentration cannot be negative")
            return
        self.self_changes = True
        with self.controller.group_action():
            self.controller.set_node_concentration(self.net_index, nodei, new_conc_float)
            post_event(DidModifyNodesEvent([nodei]))
            # TODO should have an event for this
        self.main_section.SetValidationState(True, self.conc_ctrl.GetId())

    def _OnPosText(self, evt):
        """Callback for the position control."""
        assert self.contiguous
        text = evt.GetString()
        xy = parse_num_pair(text)
        ctrl_id = self.pos_ctrl.GetId()
        if xy is None:
            self.main_section.SetValidationState(False, ctrl_id, 'Should be in the form "X, Y"')
            return

        pos = Vec2(xy)
        if pos.x < 0 or pos.y < 0:
            self.main_section.SetValidationState(
                False, ctrl_id, 'Position coordinates should be non-negative')
            return
        nodes = get_nodes_by_idx(self.all_nodes, self._selected_idx)
        # limit position to within the compartment
        compi = nodes[0].comp_idx
        if compi == -1:
            bounds = Rect(Vec2(), self.canvas.realsize)
        else:
            comp = self.canvas.comp_idx_map[compi]
            bounds = Rect(comp.position, comp.size)
        clamped = None
        index_list = list(self._selected_idx)
        if len(nodes) == 1:
            [node] = nodes
            clamped = clamp_rect_pos(Rect(pos, node.size), bounds)
            if node.position != clamped or pos != clamped:
                self.self_changes = True
                node.position = clamped
                with self.controller.group_action():
                    post_event(DidMoveNodesEvent(index_list, clamped - node.position, dragged=False))
                    self.controller.move_node(self.net_index, node.index, node.position)
        else:
            clamped = clamp_rect_pos(Rect(pos, self._bounding_rect.size), bounds)
            if self._bounding_rect.position != pos or pos != clamped:
                offset = clamped - self._bounding_rect.position
                self.self_changes = True
                with self.controller.group_action():
                    for node in nodes:
                        node.position += offset
                    post_event(DidMoveNodesEvent(index_list, offset, dragged=False))
                    for node in nodes:
                        self.controller.move_node(self.net_index, node.index, node.position)
        self.main_section.SetValidationState(True, self.pos_ctrl.GetId())

    def _OnSizeText(self, evt):
        """Callback for the size control."""
        assert self.contiguous
        ctrl_id = self.size_ctrl.GetId()
        text = evt.GetString()
        wh = parse_num_pair(text)
        if wh is None:
            self.main_section.SetValidationState(
                False, ctrl_id, 'Should be in the form "width, height"')
            return

        nodes = get_nodes_by_idx(self.all_nodes, self._selected_idx)
        min_width = get_setting('min_node_width')
        min_height = get_setting('min_node_height')
        size = Vec2(wh)
        # limit size to be smaller than the compartment
        compi = nodes[0].comp_idx
        bounds: Rect
        if compi == -1:
            bounds = Rect(Vec2(), self.canvas.realsize)
        else:
            comp = self.canvas.comp_idx_map[compi]
            bounds = comp.rect

        min_nw = min(n.size.x for n in nodes)
        min_nh = min(n.size.y for n in nodes)
        min_ratio = Vec2(min_width / min_nw, min_height / min_nh)
        min_size = self._bounding_rect.size.elem_mul(min_ratio)

        if size.x < min_size.x or size.y < min_size.y:
            message = 'The size of {} needs to be at least ({}, {})'.format(
                'bounding box' if len(nodes) > 1 else 'node',
                no_rzeros(min_size.x, 2), no_rzeros(min_size.y, 2))
            self.main_section.SetValidationState(False, ctrl_id, message)
            return

        # if size.x > max_size.x or size.y > max_size.y:
        #     message = 'The size of bounding box cannot exceed ({}, {})'.format(
        #         no_rzeros(max_size.x, 2), no_rzeros(max_size.y, 2))
        #     self.main_section.SetValidationState(False, ctrl_id, message)
        #     return

        # NOTE clamp max size automatically rather than show error
        clamped = size.reduce2(min, bounds.size)
        if self._bounding_rect.size != clamped or size != clamped:
            ratio = clamped.elem_div(self._bounding_rect.size)
            self.self_changes = True
            with self.controller.group_action():
                offsets = list()
                for node in nodes:
                    rel_pos = node.position - self._bounding_rect.position
                    new_pos = self._bounding_rect.position + rel_pos.elem_mul(ratio)
                    offsets.append(new_pos - node.position)
                    node.position = new_pos
                    node.size = node.size.elem_mul(ratio)
                    # clamp so that nodes are always within compartment/bounds
                    node.position = clamp_rect_pos(node.rect, bounds)

                idx_list = list(self._selected_idx)
                post_event(DidMoveNodesEvent(idx_list, offsets, dragged=False))
                post_event(DidResizeNodesEvent(idx_list, ratio=ratio, dragged=False))
                for node in nodes:
                    self.controller.move_node(self.net_index, node.index, node.position)
                    self.controller.set_node_size(self.net_index, node.index, node.size)
        self.main_section.SetValidationState(True, self.size_ctrl.GetId())

    def _OnNodeSizeText(self, evt):
        """Callback for the node size control."""
        # new_SBO = evt.GetString()
        # assert len(self._selected_idx) == 1
        # [nodei] = self._selected_idx
        # ctrl_SBO = self.SBO_ctrl.GetId()

        # if len(new_SBO) != 0:
        #     if new_SBO[:8] == "SBO:0000" and new_SBO[8:11].isdigit():
        #         self.self_changes = True
        #         with self.controller.group_action():
        #             self.controller.set_node_SBO(self.net_index, nodei, new_SBO)
        #             post_event(DidModifyNodesEvent([nodei]))
        #     else:
        #         self.main_section.SetValidationState(False, ctrl_SBO, "Not saved: Invalid SBO Term ID (Example: SBO:0000247)")
        #         return
         
        # self.main_section.SetValidationState(True, self.SBO_ctrl.GetId())

        assert self.contiguous
        ctrl_id = self.size_ctrl.GetId()
        text = evt.GetString()
        wh = parse_num_pair(text)

        if wh is not None: #single node selected
            nodes = get_nodes_by_idx(self.all_nodes, self._selected_idx)
            min_width = get_setting('min_node_width')
            min_height = get_setting('min_node_height')
            size = Vec2(wh)
            # limit size to be smaller than the compartment
            compi = nodes[0].comp_idx
            bounds: Rect
            if compi == -1:
                bounds = Rect(Vec2(), self.canvas.realsize)
            else:
                comp = self.canvas.comp_idx_map[compi]
                bounds = comp.rect

            min_nw = min(n.size.x for n in nodes)
            min_nh = min(n.size.y for n in nodes)
            min_ratio = Vec2(min_width / min_nw, min_height / min_nh)
            min_size = self._bounding_rect.size.elem_mul(min_ratio)

            if size.x < min_size.x or size.y < min_size.y:
                message = 'The size of {} needs to be at least ({}, {})'.format(
                    'bounding box' if len(nodes) > 1 else 'node',
                    no_rzeros(min_size.x, 2), no_rzeros(min_size.y, 2))
                self.main_section.SetValidationState(False, ctrl_id, message)
                return

            # if size.x > max_size.x or size.y > max_size.y:
            #     message = 'The size of bounding box cannot exceed ({}, {})'.format(
            #         no_rzeros(max_size.x, 2), no_rzeros(max_size.y, 2))
            #     self.main_section.SetValidationState(False, ctrl_id, message)
            #     return

            # NOTE clamp max size automatically rather than show error
            clamped = size.reduce2(min, bounds.size)
            if self._bounding_rect.size != clamped or size != clamped:
                ratio = clamped.elem_div(self._bounding_rect.size)
                self.self_changes = True
                with self.controller.group_action():
                    offsets = list()
                    for node in nodes:
                        rel_pos = node.position - self._bounding_rect.position
                        new_pos = self._bounding_rect.position + rel_pos.elem_mul(ratio)
                        offsets.append(new_pos - node.position)
                        node.position = new_pos
                        node.size = node.size.elem_mul(ratio)
                        # clamp so that nodes are always within compartment/bounds
                        node.position = clamp_rect_pos(node.rect, bounds)

                    idx_list = list(self._selected_idx)
                    post_event(DidMoveNodesEvent(idx_list, offsets, dragged=False))
                    post_event(DidResizeNodesEvent(idx_list, ratio=ratio, dragged=False))
                    for node in nodes:
                        self.controller.move_node(self.net_index, node.index, node.position)
                        self.controller.set_node_size(self.net_index, node.index, node.size)
        else: #hw is none
            nodes = get_nodes_by_idx(self.all_nodes, self._selected_idx)
            self.self_changes = True
            with self.controller.group_action():
                hws = text.split("; ")
                for i in range(len(hws)):
                    hw = hws[i][1:-1].split(", ")
                    if len(hw) != 2: #not in the format of "(w1, h1);(w2, h2)" either
                        self.main_section.SetValidationState(
                        False, ctrl_id, 'Should be in the form of "width, height" or "(w1, h1); (w2, h2)"')   
                        return
                    try:
                        self.controller.set_node_size(self.net_index, i, Vec2(float(hw[0]), float(hw[1])))
                    except:
                        self.main_section.SetValidationState(
                        False, ctrl_id, 'Should be in the form of "width, height" or "(w1, h1); (w2, h2)"')   
                        return
                   

        self.main_section.SetValidationState(True, self.size_ctrl.GetId())

    def OnNodeStatusChoice(self, evt):
        """Callback for the change node status, floating or boundary."""
        selected = self.nodeStatusDropDown.GetSelection()
        if selected == 0:
            floatingStatus = True
        else:
            floatingStatus = False

        nodes = get_nodes_by_idx(self.all_nodes, self._selected_idx)
        self.self_changes = True
        with self.controller.group_action():
            for node in nodes:
                self.controller.set_node_floating_status(self.net_index, node.index, floatingStatus)
            post_event(DidModifyNodesEvent(list(self._selected_idx)))

    def OnCompositeShapes(self, evt):
        selected = self.compositeShapesDropDown.GetStringSelection()
        nodes = get_nodes_by_idx(self.all_nodes, self._selected_idx)
        self.self_changes = True
        with self.controller.group_action():
            shapei = self.compShapeNames.index(selected)

            for node in nodes:
                if shapei != 1 and node.shape_index != 1:
                    self.controller.set_node_shape_index(self.net_index, node.index, shapei)
                else:
                    if shapei == 1: # changing node to circle
                    # if shape is being changed to circle, make bounding box sides equal
                        dim = calc_node_dimensions(node.size.x, node.size.y, 1)
                    elif node.shape_index == 1:
                        # if node was a circle previously, but not anymore, restore to default ratio
                        default_ratio = get_theme('node_height')/get_theme('node_width')
                        dim = calc_node_dimensions(node.size.x, node.size.y, default_ratio)
                    if len(nodes) == 1:
                        self.size_ctrl.SetValue(str(dim.x) + ", " + str(dim.y))
                    else:
                        self.controller.set_node_size(self.net_index, node.index, dim)
                    self.controller.set_node_shape_index(self.net_index, node.index, shapei)

            post_event(DidModifyNodesEvent(list(self._selected_idx)))

        self._UpdatePrimitiveFields()

    def OnNodeLockCheckBox(self, evt):
        """Callback for the change node lock or not."""
        cb = evt.GetEventObject()
        if cb.GetValue():
            nodeLocked = True
        else:
            nodeLocked = False

        nodes = get_nodes_by_idx(self.all_nodes, self._selected_idx)
        self.self_changes = True
        with self.controller.group_action():
            for node in nodes:
                self.controller.set_node_locked_status(self.net_index, node.index, nodeLocked)
            post_event(DidModifyNodesEvent(list(self._selected_idx)))

    def _UpdatePrimitiveFields(self):
        sizer: wx.Sizer = self.GetSizer()
        sizerflags = wx.SizerFlags().Expand()

        self.Freeze()

        nodes = self.selected_nodes
        shape_names = set(n.composite_shape.name for n in nodes)

        if self.last_prim_section is not None:
            sizer.Detach(self.last_prim_section)
            self.last_prim_section.Hide()

        if len(shape_names) == 1:
            shape_index = nodes[0].shape_index

            if shape_index in self.prim_section_cache:
                # already created form for this shape before; restore the cached one
                prim_section = self.prim_section_cache[shape_index]
                prim_section.Show()
                # self.AddChild(prim_section)
                sizer.Add(prim_section, sizerflags)
            else:
                # need to create new one
                assert nodes[0].composite_shape is not None
                prim_section = PrimitiveSection(self, nodes[0].composite_shape)
                sizer.Add(prim_section, sizerflags)
                self.prim_section_cache[shape_index] = prim_section

            self.last_prim_section = prim_section

            prim_section.UpdatePrimitiveValues()
        else:
            pass  # don't need to do anything since this whole section is hidden

        # need to tell parent to adjust height as well
        self.GetParent().Layout()
        self.Thaw()

    def UpdateAllFields(self):
        """Update the form field values based on current data."""
        self.self_changes = False
        assert len(self._selected_idx) != 0
        nodes = get_nodes_by_idx(self.all_nodes, self._selected_idx)
        prec = get_setting('decimal_precision')
        id_text: str
        conc_text: str
        floatingNode: bool
        lockNode: bool
        shape_name: str

        if not self.contiguous:
            self.pos_ctrl.ChangeValue('?')
            #self.size_ctrl.ChangeValue('?')
        else:
            ChangePairValue(self.pos_ctrl, self._bounding_rect.position, prec)
            #ChangePairValue(self.size_ctrl, self._bounding_rect.size, prec)

        if not self.contiguous:
            self.size_ctrl.ChangeValue('?')

        if len(self._selected_idx) == 1:
            [node] = nodes
            self.id_ctrl.Enable(True)
            id_text = node.id
            self.comp_ctrl.Enable(False)
            #comp_text = str(node.comp_idx)
            allCompartments = api.get_compartments(self.net_index)
            comp_id = ''
            for comp in allCompartments:
                if comp.index == node.comp_idx:
                    comp_id = comp.id
            comp_text = str(comp_id)
            self.name_ctrl.Enable(True)
            name_text = str(node.node_name)
            self.SBO_ctrl.Enable(True)
            SBO_text = str(node.node_SBO)
            self.conc_ctrl.Enable(True)
            conc_text = str(node.concentration)
            self.size_ctrl.Enable(True)
            size_text = str(node.size[0]) + ", " + str(node.size[1])
            floatingNode = node.floatingNode
            lockNode = node.lockNode
            assert node.composite_shape is not None
            shape_name = node.composite_shape.name
        else:
            self.id_ctrl.Enable(False)
            id_text = '; '.join(sorted(list(n.id for n in nodes)))
            self.comp_ctrl.Enable(False)
            allCompartments = api.get_compartments(self.net_index)
            comp_list = []
            for n in nodes:
                comp_id = ''
                for comp in allCompartments:
                    if comp.index == n.comp_idx:
                        comp_id = comp.id
                comp_list.append(str(comp_id))
            comp_text = '; '.join(sorted(comp_list))
            self.name_ctrl.Enable(False)
            name_text = '; '.join(sorted(list(str(n.node_name) for n in nodes)))
            self.SBO_ctrl.Enable(False)
            SBO_text = '; '.join(sorted(list(str(n.node_SBO) for n in nodes)))
            self.conc_ctrl.Enable(False)
            conc_text = '; '.join(sorted(list(str(n.concentration) for n in nodes)))
            self.size_ctrl.Enable(True)
            size_text = '; '.join(sorted(list(str(n.size) for n in nodes)))
            floatingNode = all(n.floatingNode for n in nodes)
            lockNode = all(n.lockNode for n in nodes)
            shape_name_set = set(n.composite_shape.name for n in nodes)
            if len(shape_name_set) == 1:
                shape_name = next(iter(shape_name_set))
            else:
                shape_name = ''

        self._UpdatePrimitiveFields()
        self.pos_ctrl.Enable(self.contiguous)
        self.size_ctrl.Enable(self.contiguous)

        self.id_ctrl.ChangeValue(id_text)
        self.comp_ctrl.ChangeValue(comp_text)
        self.name_ctrl.ChangeValue(name_text)
        self.SBO_ctrl.ChangeValue(SBO_text)
        self.conc_ctrl.ChangeValue(conc_text)
        self.size_ctrl.ChangeValue(size_text)

        if floatingNode:
            self.nodeStatusDropDown.SetSelection(0)
        else:
            self.nodeStatusDropDown.SetSelection(1)

        if lockNode:
            self.lockNodeCheckBox.SetValue(True)
        else:
            self.lockNodeCheckBox.SetValue(False)

        if shape_name:
            self.compositeShapesDropDown.SetStringSelection(shape_name)
        else:
            self.compositeShapesDropDown.SetSelection(wx.NOT_FOUND)


@dataclass
class StoichInfo:
    """Helper class that stores node stoichiometry info for reaction form"""
    nodei: int
    stoich: float


'''Section for editing stoichiometry, includes only reactants or only products,
so there are two of this.'''
class StoichSection(FieldGrid):
    def __init__(self, parent, form: 'ReactionForm', stoichs: List[StoichInfo], reai: int,
                 is_reactants: bool):
        super().__init__(parent, form)

        self._reactant_subtitle = self.AppendSubtitle('Reactants' if is_reactants else 'Products')
        for stoich in stoichs:
            stoich_ctrl = self.CreateTextCtrl(value=no_rzeros(stoich.stoich, precision=2))
            node_id = self.form.controller.get_node_id(self.form.net_index, stoich.nodei)
            self.AppendControl(node_id, stoich_ctrl)
            if is_reactants:
                inner_callback = self.MakeSetSrcStoichFunction(reai, stoich.nodei)
            else:
                inner_callback = self.MakeSetDestStoichFunction(reai, stoich.nodei)
            callback = self.MakeFloatCtrlFunction(stoich_ctrl.GetId(), inner_callback, (0, None),
                                                    left_incl=False)
            stoich_ctrl.Bind(wx.EVT_TEXT, callback)

    def MakeSetSrcStoichFunction(self, reai: int, nodei: int):
        def ret(val: float):
            self.form.self_changes = True
            with self.form.controller.group_action():
                self.form.controller.set_src_node_stoich(self.form.net_index, reai, nodei, val)
                post_event(DidModifyReactionEvent(list(self.form.selected_idx)))

        return ret

    def MakeSetDestStoichFunction(self, reai: int, nodei: int):
        def ret(val: float):
            with self.form.controller.group_action():
                self.form.self_changes = True
                self.form.controller.set_dest_node_stoich(self.form.net_index, reai, nodei, val)
                post_event(DidModifyReactionEvent(list(self.form.selected_idx)))

        return ret


class ReactionForm(EditPanelForm):
    auto_center_ctrl: wx.CheckBox

    def __init__(self, parent, canvas: Canvas, controller: IController):
        super().__init__(parent, canvas, controller)

        self.reactions = list()
        self.main_section = FieldGrid(self, self)
        self.sections = [self.main_section]

        self.CreateChildren()

    def CreateControls(self):
        self.id_ctrl = self.main_section.CreateTextCtrl()
        self.id_ctrl.Bind(wx.EVT_TEXT, self._OnIdText)
        self.main_section.AppendControl('identifier', self.id_ctrl)

        self.ratelaw_ctrl = self.main_section.CreateTextCtrl()
        #self.ratelaw_ctrl.Bind(wx.EVT_TEXT, self._OnRateLawText)
        self.main_section.AppendControl('rate law', self.ratelaw_ctrl)

        self.fill_ctrl, self.fill_alpha_ctrl = self.main_section.CreateColorControl(
            'fill color', 'fill opacity',
            self._OnFillColorChanged, self._FillAlphaCallback)

        self.stroke_width_ctrl = self.main_section.CreateTextCtrl()
        stroke_cb = self.main_section.MakeFloatCtrlFunction(self.stroke_width_ctrl.GetId(),
                                                            self._StrokeWidthCallback, (0.1, 100))
        self.stroke_width_ctrl.Bind(wx.EVT_TEXT, stroke_cb)
        self.main_section.AppendControl('line width', self.stroke_width_ctrl)

        # Whether the center position should be autoly set?
        #self.auto_center_ctrl = wx.CheckBox(self.main_section)
        #self.auto_center_ctrl.SetValue(True)
        #self.auto_center_ctrl = wx.CheckBox(self.main_section, label = '')


        # self.auto_center_ctrl = wx.CheckBox(self.main_section, label = '')
        # self.auto_center_ctrl.SetValue(False)
        # self.auto_center_ctrl.Bind(wx.EVT_CHECKBOX, self._AutoCenterCallback)
        # self.main_section.AppendControl('auto center pos', self.auto_center_ctrl)


        #self.auto_center_ctrl = wx.CheckBox(self.main_section, label = '')
        self.auto_center_ctrl = wx.ToggleButton(self.main_section, -1, '')
        self.auto_center_ctrl.SetValue(False)
        self.auto_center_ctrl.SetLabel("Off")
        self.auto_center_ctrl.Bind(wx.EVT_TOGGLEBUTTON, self._AutoCenterCallback)
        self.main_section.AppendControl('auto center', self.auto_center_ctrl)


        self.center_pos_ctrl = self.main_section.CreateTextCtrl()
        self.center_pos_ctrl.Disable()
        self.center_pos_ctrl.Bind(wx.EVT_TEXT, self._CenterPosCallback)
        self.main_section.AppendControl('center position', self.center_pos_ctrl)

        self._reactant_subtitle = None
        self._product_subtitle = None
        self.reactant_stoich_ctrls = list()
        self.product_stoich_ctrls = list()

        states = ['bezier curve', 'straight line']
        self.rxnStatusDropDown = wx.Choice(self.main_section, choices=states)
        self.main_section.AppendControl('reaction status', self.rxnStatusDropDown)
        self.rxnStatusDropDown.Bind(wx.EVT_CHOICE, self.OnRxnStatusChoice)

        self.mod_tip_dropdown = wx.ComboBox(
            self.main_section, choices=[e.value for e in ModifierTipStyle], style=wx.CB_READONLY)
        self.main_section.AppendControl('modifier tip', self.mod_tip_dropdown)
        self.mod_tip_dropdown.Bind(wx.EVT_COMBOBOX, self.ModifierTipCallback)

        self._modifiers = set()
        self.all_nodes = list()
        self.node_indices = set()
        self.modifiers_ctrl = wx.CheckListBox(
            self.main_section, style=wx.LB_NEEDED_SB, size=(-1, 100))
        self.main_section.AppendControl('modifiers', self.modifiers_ctrl)
        self.modifiers_ctrl.Bind(wx.EVT_CHECKLISTBOX, self.OnModifierCheck)
        self.reactants_section = None
        self.products_section = None

    def _OnIdText(self, evt):
        """Callback for the ID control."""
        new_id = evt.GetString()
        assert len(self._selected_idx) == 1, 'Reaction ID field should be disabled when ' + \
            'multiple are selected'
        [reai] = self._selected_idx
        ctrl_id = self.id_ctrl.GetId()
        if len(new_id) == 0:
            self.main_section.SetValidationState(False, ctrl_id, "ID cannot be empty")
            return
        else:
            for rxn in self.reactions:
                if rxn.id == new_id:
                    self.main_section.SetValidationState(False, ctrl_id, "Not saved: Duplicate ID")
                    return

            # loop terminated fine. There is no duplicate ID
            self.self_changes = True
            with self.controller.group_action():
                self.controller.rename_reaction(self.net_index, reai, new_id)
                post_event(DidModifyReactionEvent(list(self._selected_idx)))
            self.main_section.SetValidationState(True, ctrl_id)

    def _StrokeWidthCallback(self, width: float):
        reactions = [r for r in self.reactions if r.index in self._selected_idx]
        self.self_changes = True
        with self.controller.group_action():
            for rxn in reactions:
                self.controller.set_reaction_line_thickness(self.net_index, rxn.index, width)
            post_event(DidModifyReactionEvent(list(self._selected_idx)))

    def OnRxnStatusChoice(self, evt):
        """Callback for the change reaction status, bezier curve or straight line."""
        selection = self.rxnStatusDropDown.GetSelection()
        # TODO this is hardcoded. If the text changes this wouldn't work
        if selection == 0:
            bezierCurves = True
        else:
            bezierCurves = False

        rxns = get_rxns_by_idx(self.reactions, self._selected_idx)
        self.self_changes = True
        with self.controller.group_action():
            for rxn in rxns:
                self.controller.set_reaction_bezier_curves(self.net_index, rxn.index, bezierCurves)
            post_event(DidModifyReactionEvent(list(self._selected_idx)))

    def ModifierTipCallback(self, evt):
        """Callback for the change reaction status, bezier curve or straight line."""
        status = self.mod_tip_dropdown.GetValue()
        entry: ModifierTipStyle
        for e in ModifierTipStyle:
            if e.value == status:
                entry = e
                break
        else:
            assert False, ('Unable to find corresponding enum entry to dropdown selection. ' +
                           'This is not supposed to happen.')

        rxns = get_rxns_by_idx(self.reactions, self._selected_idx)
        self.self_changes = True
        with self.controller.group_action():
            for rxn in rxns:
                self.controller.set_modifier_tip_style(self.net_index, rxn.index, entry)
            post_event(DidModifyReactionEvent(list(self._selected_idx)))

    def _AutoCenterCallback(self, evt):

        checked = evt.GetInt()

        assert len(self._selected_idx) == 1
        prec = 2
        reaction = self.canvas.reaction_idx_map[next(iter(self._selected_idx))]
        centroid_map = self.canvas.GetReactionCentroids(self.net_index)
        centroid = centroid_map[reaction.index]
        if checked:
            self.auto_center_ctrl.Enable()
            self.auto_center_ctrl.SetValue(True)
            self.auto_center_ctrl.SetLabel("On")
            self.center_pos_ctrl.ChangeValue('')
            self.center_pos_ctrl.Disable()
            with self.controller.group_action():
                self.controller.set_reaction_center(self.net_index, reaction.index, None)
                # Move centroid handle along if centroid changed.
                if reaction.center_pos is not None:
                    offset = centroid - reaction.center_pos
                    if offset != Vec2():
                        self.controller.set_center_handle(
                            self.net_index, reaction.index, reaction.src_c_handle.tip + offset)
            self.center_pos_ctrl.Disable()
        else:
            self.auto_center_ctrl.Enable()
            self.auto_center_ctrl.SetValue(False)
            self.auto_center_ctrl.SetLabel("Off")
            self.center_pos_ctrl.Enable()
            self.center_pos_ctrl.ChangeValue('{}, {}'.format(
                no_rzeros(centroid.x, prec), no_rzeros(centroid.y, prec)
            ))
            self.controller.set_reaction_center(self.net_index, reaction.index, centroid)
        

    def _CenterPosCallback(self, evt):
        text = evt.GetString()
        xy = parse_num_pair(text)
        ctrl_id = self.center_pos_ctrl.GetId()
        if xy is None:
            self.main_section.SetValidationState(False, ctrl_id, 'Should be in the form "X, Y"')
            return

        pos = Vec2(xy)
        if pos.x < 0 or pos.y < 0:
            self.main_section.SetValidationState(
                False, ctrl_id, 'Position coordinates should be non-negative')
            return

        assert len(self._selected_idx) == 1
        reaction = self.canvas.reaction_idx_map[next(iter(self._selected_idx))]
        if reaction.center_pos != pos:
            offset = pos - reaction.center_pos
            self.self_changes = True
            with self.controller.group_action():
                self.controller.set_reaction_center(self.net_index, reaction.index, pos)
                post_event(DidMoveReactionCenterEvent(self.net_index, reaction.index, offset, False))
        self.main_section.SetValidationState(True, ctrl_id)

    def _OnFillColorChanged(self, fill: wx.Colour):
        """Callback for the fill color control."""
        reactions = [r for r in self.reactions if r.index in self._selected_idx]
        self.self_changes = True
        with self.controller.group_action():
            for rxn in reactions:
                if on_msw():
                    self.controller.set_reaction_fill_rgb(self.net_index, rxn.index, fill)
                else:
                    # we can set both the RGB and the alpha at the same time
                    self.controller.set_reaction_fill_rgb(self.net_index, rxn.index, fill)
                    self.controller.set_reaction_fill_alpha(self.net_index, rxn.index, fill.Alpha())
            post_event(DidModifyReactionEvent(list(self._selected_idx)))

    def _FillAlphaCallback(self, alpha: float):
        """Callback for when the fill alpha changes."""
        reactions = (r for r in self.reactions if r.index in self._selected_idx)
        self.self_changes = True
        with self.controller.group_action():
            for rxn in reactions:
                self.controller.set_reaction_fill_alpha(self.net_index, rxn.index, int(alpha * 255))
            post_event(DidModifyReactionEvent(list(self._selected_idx)))

    def _OnRateLawText(self, evt: wx.CommandEvent):
        ratelaw = evt.GetString()
        assert len(self._selected_idx) == 1, 'Reaction rate law field should be disabled when ' + \
            'multiple are selected'
        [reai] = self._selected_idx
        self.self_changes = True
        post_event(DidModifyReactionEvent(list(self._selected_idx)))
        self.controller.set_reaction_ratelaw(self.net_index, reai, ratelaw)

    def OnModifierCheck(self, evt: wx.CommandEvent):
        evt.Skip()
        assert len(self._selected_idx) == 1
        reactions = [r for r in self.reactions if r.index in self._selected_idx]
        assert len(reactions) == 1
        reaction = reactions[0]
        new_modifiers = [self.all_nodes[i].index for i in self.modifiers_ctrl.GetCheckedItems()]
        self.controller.set_reaction_modifiers(self.net_index, reaction.index, new_modifiers)

    def CanvasUpdated(self, reactions: List[Reaction], nodes: List[Node]):
        """Function called after the canvas has been updated."""
        self.reactions = reactions
        self.all_nodes = nodes
        new_node_indices = set(n.index for n in nodes)
        # if new_node_indices != self.node_indices:
        self.node_indices = new_node_indices
        self._UpdateModifierList()
        self.ExternalUpdate()

    def UpdateSelection(self, selected_idx: List[int]):
        """Function called after the list of selected reactions have been updated."""
        self._selected_idx = selected_idx
        if len(self._selected_idx) != 0:
            title_label = 'Edit Reaction' if len(self._selected_idx) == 1 \
                else 'Edit Multiple Reactions'
            self._title.SetLabel(title_label)

            id_text = 'identifier' if len(self._selected_idx) == 1 else 'identifiers'
            self.main_section.labels[self.id_ctrl.GetId()].SetLabel(id_text)
            self.UpdateAllFields()
        self.ExternalUpdate()

    def _UpdateModifierList(self):
        # NOTE if slightly better performance is wanted, we don't have to update this widget
        # immediately. Rather we can have a dirty flag and update only when displaying
        node_names = list()
        for n in self.all_nodes:
            name = n.id
            if n.original_index != -1:
                name += ' (alias)'
            node_names.append(name)

        self.modifiers_ctrl.Set(node_names)
        self._UpdateModifierSelection()

    def _UpdateModifierSelection(self):
        checked_indices = set(i for i, n in enumerate(self.all_nodes) if n.index in self._modifiers)
        self.modifiers_ctrl.SetCheckedItems(checked_indices)

    def _UpdateStoichFields(self, reai: int, reactants: List[StoichInfo], products: List[StoichInfo]):
        sizer = self.GetSizer()
        sizerflags = wx.SizerFlags().Expand()

        self.Freeze()
        if self.reactants_section is not None:
            sizer.Detach(self.reactants_section)
            sizer.Detach(self.products_section)
            self.reactants_section.Destroy()
            self.products_section.Destroy()

        if len(reactants) != 0:
            assert len(products) != 0
            self.reactants_section = StoichSection(self, self, reactants, reai, True)
            self.products_section = StoichSection(self, self, products, reai, False)
            sizer.Add(self.reactants_section, sizerflags)
            sizer.Add(self.products_section, sizerflags)
        else:
            self.reactants_section = None
            self.products_section = None
            # Both reactants and products are empty; don't add
            assert len(products) == 0

        self.Layout()
        self.Thaw()

    def _GetSrcStoichs(self, reai: int):
        ids = self.controller.get_list_of_src_indices(self.net_index, reai)
        return [StoichInfo(id, self.controller.get_src_node_stoich(self.net_index, reai, id))
                for id in ids]

    def _GetDestStoichs(self, reai: int):
        ids = self.controller.get_list_of_dest_indices(self.net_index, reai)
        return [StoichInfo(id, self.controller.get_dest_node_stoich(self.net_index, reai, id))
                for id in ids]

    def UpdateAllFields(self):
        """Update all reaction fields from current data."""
        self.self_changes = False
        assert len(self._selected_idx) != 0
        reactions = [r for r in self.reactions if r.index in self._selected_idx]
        id_text = '; '.join(sorted(list(r.id for r in reactions)))
        fill: wx.Colour
        fill_alpha: Optional[int]
        ratelaw_text: str

        prec = get_setting('decimal_precision')

        if len(self._selected_idx) == 1:
            [reaction] = reactions
            reai = reaction.index
            self.id_ctrl.Enable()
            fill = reaction.fill_color
            fill_alpha = reaction.fill_color.Alpha()
            ratelaw_text = reaction.rate_law
            #self.ratelaw_ctrl.Enable()
            self.ratelaw_ctrl.Disable()
            
            self.auto_center_ctrl.Enable()
            auto_set = reaction.center_pos is None
            self.auto_center_ctrl.SetValue(auto_set)
            if auto_set:
                self.auto_center_ctrl.SetLabel("On")
                self.center_pos_ctrl.ChangeValue(' ')
            else:
                self.auto_center_ctrl.SetLabel("Off")
            self.center_pos_ctrl.Enable(not auto_set)
            if reaction.center_pos is not None:
                centroid = reaction.center_pos
                self.center_pos_ctrl.ChangeValue('{}, {}'.format(
                    no_rzeros(centroid.x, prec), no_rzeros(centroid.y, prec)
                ))

            self._UpdateStoichFields(reai, self._GetSrcStoichs(reai), self._GetDestStoichs(reai))
            self.modifiers_ctrl.Enable()
            self._modifiers = reaction.modifiers
            self._UpdateModifierSelection()
        else:

            self.id_ctrl.Disable()
            fill, fill_alpha = GetMultiColor(list(r.fill_color for r in reactions))
            ratelaw_text = 'multiple'
            self.ratelaw_ctrl.Disable()
            self.auto_center_ctrl.SetValue(False)
            self.auto_center_ctrl.SetLabel("Off")
            self.auto_center_ctrl.Disable()
            self.center_pos_ctrl.ChangeValue(' ')
            self.center_pos_ctrl.Disable()

            self._UpdateStoichFields(0, [], [])
            self.modifiers_ctrl.Disable()
            self._modifiers = set()
            self._UpdateModifierSelection()

        bezierCurves = all(r.bezierCurves for r in reactions)
        mod_tip_style = GetMultiEnum(
            list(r.modifier_tip_style for r in reactions), ModifierTipStyle.CIRCLE)
        stroke_width = GetMultiFloatText(set(r.thickness for r in reactions), prec)

        self.id_ctrl.ChangeValue(id_text)
        self.fill_ctrl.SetColour(fill)
        self.ratelaw_ctrl.ChangeValue(ratelaw_text)
        self.stroke_width_ctrl.ChangeValue(stroke_width)
        self.mod_tip_dropdown.SetValue(mod_tip_style.value)

        if on_msw():
            self.fill_alpha_ctrl.ChangeValue(AlphaToText(fill_alpha, prec))

        # HMS Replaced stings with contants
        if bezierCurves:
            self.rxnStatusDropDown.SetSelection(0)
        else:
            self.rxnStatusDropDown.SetSelection(1)


class CompartmentForm(EditPanelForm):
    _compartments: List[Compartment]
    contiguous: bool

    def __init__(self, parent, canvas: Canvas, controller: IController):
        super().__init__(parent, canvas, controller)
        self.compartments = list()
        self.contiguous = True

        self.main_section = FieldGrid(self, self)
        self.sections = [self.main_section]

        self.CreateChildren()

    def CreateControls(self):
        self.id_ctrl = self.main_section.CreateTextCtrl()
        self.id_ctrl.Bind(wx.EVT_TEXT, self._OnIdText)
        self.main_section.AppendControl('identifier', self.id_ctrl)

        self.pos_ctrl = self.main_section.CreateTextCtrl()
        self.pos_ctrl.Bind(wx.EVT_TEXT, self._OnPosText)
        self.main_section.AppendControl('position', self.pos_ctrl)

        self.size_ctrl = self.main_section.CreateTextCtrl()
        self.size_ctrl.Bind(wx.EVT_TEXT, self._OnSizeText)
        self.main_section.AppendControl('size', self.size_ctrl)

        self.volume_ctrl = self.main_section.CreateTextCtrl()
        self.main_section.AppendControl('volume', self.volume_ctrl)
        volume_callback = self.main_section.MakeFloatCtrlFunction(self.volume_ctrl.GetId(),
                                                                  self._VolumeCallback, (0, None), left_incl=False)
        self.volume_ctrl.Bind(wx.EVT_TEXT, volume_callback)

        self.fill_ctrl, self.fill_alpha_ctrl = self.main_section.CreateColorControl(
            'fill color', 'fill opacity',
            self._OnFillColorChanged, self._FillAlphaCallback,)

        self.border_ctrl, self.border_alpha_ctrl = self.main_section.CreateColorControl(
            'border color', 'border opacity',
            self._OnBorderColorChanged, self._BorderAlphaCallback)

        self.border_width_ctrl = self.main_section.CreateTextCtrl()
        self.main_section.AppendControl('border width', self.border_width_ctrl)
        border_callback = self.main_section.MakeFloatCtrlFunction(self.border_width_ctrl.GetId(),
                                                                  self._BorderWidthCallback, (1, 100))
        self.border_width_ctrl.Bind(wx.EVT_TEXT, border_callback)

    def _OnIdText(self, evt):
        """Callback for the ID control."""
        new_id = evt.GetString()
        assert len(self._selected_idx) == 1, 'Compartment ID field should be disabled when ' + \
            'multiple are selected'
        [compi] = self._selected_idx
        ctrl_id = self.id_ctrl.GetId()
        if len(new_id) == 0:
            self.main_section.SetValidationState(False, ctrl_id, "ID cannot be empty")
            return
        else:
            for comp in self.compartments:
                if comp.id == new_id:
                    self.main_section.SetValidationState(False, ctrl_id, "Not saved: Duplicate ID")
                    return

            # loop terminated fine. There is no duplicate ID
            self.self_changes = True
            with self.controller.group_action():
                self.controller.rename_compartment(self.net_index, compi, new_id)
                post_event(DidModifyCompartmentsEvent(list(self._selected_idx)))
            self.main_section.SetValidationState(True, ctrl_id)

    def _OnPosText(self, evt):
        """Callback for the position control."""
        assert self.contiguous
        text = evt.GetString()
        xy = parse_num_pair(text)
        ctrl_id = self.pos_ctrl.GetId()
        if xy is None:
            self.main_section.SetValidationState(False, ctrl_id, 'Should be in the form "X, Y"')
            return

        pos = Vec2(xy)
        if pos.x < 0 or pos.y < 0:
            self.main_section.SetValidationState(
                False, ctrl_id, 'Position coordinates should be non-negative')
            return
        comps = [c for c in self.compartments if c.index in self._selected_idx]
        bounds = Rect(Vec2(), self.canvas.realsize)
        clamped = clamp_rect_pos(Rect(pos, self._bounding_rect.size), bounds)
        if self._bounding_rect.position != pos or pos != clamped:
            offset = clamped - self._bounding_rect.position
            self.self_changes = True
            with self.controller.group_action():
                for comp in comps:
                    comp.position += offset
                post_event(DidMoveCompartmentsEvent(list(self._selected_idx), offset, dragged=False))
                for comp in comps:
                    self.controller.move_node(self.net_index, comp.index, comp.position)
        self.main_section.SetValidationState(True, self.pos_ctrl.GetId())

    def _OnSizeText(self, evt):
        """Callback for the size control."""
        ctrl_id = self.size_ctrl.GetId()
        text = evt.GetString()
        wh = parse_num_pair(text)
        if wh is None:
            self.main_section.SetValidationState(
                False, ctrl_id, 'Should be in the form "width, height"')
            return

        comps = [c for c in self.compartments if c.index in self._selected_idx]
        size = Vec2(wh)
        _, comp_min_ratio = self.canvas.select_box.compute_min_ratio()
        assert comp_min_ratio is not None
        limit = self._bounding_rect.size.elem_mul(comp_min_ratio)

        if size.x < limit.x or size.y < limit.y:
            message = 'Size of {} needs to be at least ({}, {})'.format(
                'bounding box' if len(comps) > 1 else 'compartment',
                no_rzeros(limit.x, 2), no_rzeros(limit.y, 2))
            self.main_section.SetValidationState(False, ctrl_id, message)
            return

        clamped = clamp_rect_size(Rect(self._bounding_rect.position, size), self.canvas.realsize)
        if self._bounding_rect.size != clamped or size != clamped:
            ratio = clamped.elem_div(self._bounding_rect.size)
            self.self_changes = True
            with self.controller.group_action():
                offsets = list()
                peripheral_nodes = list()
                peripheral_offsets = list()
                for comp in comps:
                    rel_pos = comp.position - self._bounding_rect.position
                    new_pos = self._bounding_rect.position + rel_pos.elem_mul(ratio)
                    offsets.append(new_pos - comp.position)
                    comp.position = new_pos
                    comp.size = comp.size.elem_mul(ratio)

                    pnodes = [self.canvas.node_idx_map[i] for i in comp.nodes]
                    for node in pnodes:
                        new_pos = clamp_rect_pos(node.rect, comp.rect)
                        if new_pos != node.position:
                            node.position = new_pos
                            peripheral_nodes.append(node)
                            peripheral_offsets.append(new_pos - node.position)

                idx_list = list(self._selected_idx)
                post_event(DidMoveCompartmentsEvent(idx_list, offsets, dragged=False))
                post_event(DidResizeCompartmentsEvent(idx_list, ratio, dragged=False))
                if len(peripheral_nodes) != 0:
                    post_event(DidMoveNodesEvent(peripheral_nodes, peripheral_offsets, dragged=False))
                for comp in comps:
                    self.controller.move_compartment(self.net_index, comp.index, comp.position)
                    self.controller.set_compartment_size(self.net_index, comp.index, comp.size)

                for node in peripheral_nodes:
                    self.controller.move_node(self.net_index, node.index, node.position)
        self.main_section.SetValidationState(True, self.size_ctrl.GetId())

    def _VolumeCallback(self, volume: float):
        """Callback for when the border width changes."""
        comps = [c for c in self.compartments if c.index in self._selected_idx]
        self.self_changes = True
        with self.controller.group_action():
            for comp in comps:
                self.controller.set_compartment_volume(self.net_index, comp.index, volume)
            post_event(DidModifyCompartmentsEvent(list(self._selected_idx)))

    def _OnFillColorChanged(self, fill: wx.Colour):
        """Callback for the fill color control."""
        comps = [c for c in self.compartments if c.index in self._selected_idx]
        self.self_changes = True
        with self.controller.group_action():
            for comp in comps:
                if on_msw():
                    fill = wx.Colour(fill.GetRGB())  # remove alpha channel
                self.controller.set_compartment_fill(self.net_index, comp.index, fill)
            post_event(DidModifyCompartmentsEvent(list(self._selected_idx)))

    def _OnBorderColorChanged(self, border: wx.Colour):
        """Callback for the border color control."""
        comps = [c for c in self.compartments if c.index in self._selected_idx]
        self.self_changes = True
        with self.controller.group_action():
            for comp in comps:
                if on_msw():
                    border = wx.Colour(border.GetRGB())  # remove alpha channel
                self.controller.set_compartment_border(self.net_index, comp.index, border)
            post_event(DidModifyCompartmentsEvent(list(self._selected_idx)))

    def _FillAlphaCallback(self, alpha: float):
        """Callback for when the fill alpha changes."""
        comps = [c for c in self.compartments if c.index in self._selected_idx]
        self.self_changes = True
        with self.controller.group_action():
            for comp in comps:
                new_fill = change_opacity(comp.fill, int(alpha * 255))
                self.controller.set_compartment_fill(self.net_index, comp.index, new_fill)
            post_event(DidModifyCompartmentsEvent(list(self._selected_idx)))

    def _BorderAlphaCallback(self, alpha: float):
        """Callback for when the border alpha changes."""
        comps = [c for c in self.compartments if c.index in self._selected_idx]
        self.self_changes = True
        with self.controller.group_action():
            for comp in comps:
                new_border = change_opacity(comp.border, int(alpha * 255))
                self.controller.set_compartment_border(self.net_index, comp.index, new_border)
            post_event(DidModifyCompartmentsEvent(list(self._selected_idx)))

    def _BorderWidthCallback(self, width: float):
        """Callback for when the border width changes."""
        comps = [c for c in self.compartments if c.index in self._selected_idx]
        self.self_changes = True
        with self.controller.group_action():
            for comp in comps:
                self.controller.set_compartment_border_width(self.net_index, comp.index, width)
            post_event(DidModifyCompartmentsEvent(list(self._selected_idx)))

    def UpdateCompartments(self, comps: List[Compartment]):
        self.compartments = comps
        self._UpdateBoundingRect()
        self.ExternalUpdate()

    def UpdateSelection(self, selected_idx: List[int], nodes_selected: bool):
        self._selected_idx = selected_idx
        if len(selected_idx) != 0:
            self._UpdateBoundingRect()

        self.contiguous = not nodes_selected

        if len(self._selected_idx) != 0:
            # clear position value
            # self.pos_ctrl.ChangeValue('')
            self.UpdateAllFields()

            title_label = 'Edit Compartment' if len(
                self._selected_idx) == 1 else 'Edit Multiple Compartments'
            self._title.SetLabel(title_label)

            id_text = 'identifier' if len(self._selected_idx) == 1 else 'identifiers'
            self.main_section.labels[self.id_ctrl.GetId()].SetLabel(id_text)
        self.ExternalUpdate()

    def UpdateAllFields(self):
        self.self_changes = False
        comps = [c for c in self.compartments if c.index in self._selected_idx]
        assert len(comps) == len(self._selected_idx)
        prec = 2

        id_text = '; '.join([c.id for c in comps])
        fill: wx.Colour
        fill_alpha: Optional[int]
        border: wx.Colour

        self.pos_ctrl.Enable(self.contiguous)
        self.size_ctrl.Enable(self.contiguous)
        border_width = GetMultiFloatText(set(c.border_width for c in comps), prec)
        volume = GetMultiFloatText(set(c.volume for c in comps), prec)

        if not self.contiguous:
            self.pos_ctrl.ChangeValue('?')
            self.size_ctrl.ChangeValue('?')
        else:
            ChangePairValue(self.pos_ctrl, self._bounding_rect.position, prec)
            ChangePairValue(self.size_ctrl, self._bounding_rect.size, prec)

        if len(self._selected_idx) == 1:
            [comp] = comps
            self.id_ctrl.Enable()
            fill = comp.fill
            fill_alpha = comp.fill.Alpha()
            border = comp.border
            border_alpha = comp.border.Alpha()
        else:
            self.id_ctrl.Disable()
            fill, fill_alpha = GetMultiColor(list(c.fill for c in comps))
            border, border_alpha = GetMultiColor(list(c.border for c in comps))

        self.id_ctrl.ChangeValue(id_text)
        self.fill_ctrl.SetColour(fill)
        self.border_ctrl.SetColour(border)
        self.volume_ctrl.ChangeValue(volume)

        # set fill alpha if on windows
        if on_msw():
            self.fill_alpha_ctrl.ChangeValue(AlphaToText(fill_alpha, prec))
            self.border_alpha_ctrl.ChangeValue(AlphaToText(border_alpha, prec))

        self.border_width_ctrl.ChangeValue(border_width)

    def CompsMovedOrResized(self, evt):
        """Called when nodes are moved or resized by dragging"""
        if not evt.dragged:
            return
        self._UpdateBoundingRect()
        prec = 2
        ChangePairValue(self.pos_ctrl, self._bounding_rect.position, prec)
        ChangePairValue(self.size_ctrl, self._bounding_rect.size, prec)

    def _UpdateBoundingRect(self):
        """Update bounding rectangle; mixed indicates whether both nodes and comps are selected.
        """
        rects = [c.rect for c in self.compartments if c.index in self._selected_idx]
        # It could be that compartments have been updated but selected indices have not.
        # In that case rects can be empty
        if len(rects) != 0:
            self._bounding_rect = get_bounding_rect(rects)



"""Classes for storing and managing data for graph elements."""
# from __future__ import annotations
# pylint: disable=maybe-no-member
from dataclasses import dataclass
from enum import Enum, auto
from functools import reduce
import wx
import copy
import math
from math import factorial, pi, sin, cos
from operator import mul
from itertools import chain
import numpy as np
from typing import Any, Callable, ClassVar, Container, List, NamedTuple, Optional, Sequence, Set, Tuple, cast
from collections import namedtuple

from .geometry import Vec2, Rect, get_bounding_rect, padded_rect, pt_in_circle, pt_on_line, rotate_unit, segment_rect_intersection, segments_intersect
from .state import cstate
from ..config import get_setting, get_theme, Color, Font
from ..utils import gchain, pairwise


MAXSEGS = 8  # Number of segments used to construct bezier
HANDLE_RADIUS = 5  # Radius of the contro lhandle
HANDLE_BUFFER = 2
NODE_EDGE_GAP_DISTANCE = 4  # Distance between node and start of bezier line
TIP_DISPLACEMENT = 4


@dataclass(frozen=True)
class Transform:
    translation: Vec2 = Vec2()
    rotation: float = 0
    scale: Vec2 = Vec2(1, 1)


@dataclass()
class Primitive:
    name: ClassVar[str] = 'generic primitive'


@dataclass()
class CirclePrim(Primitive):
    name: ClassVar[str] = 'circle'
    fill_color: Color = Color(255, 0, 0, 255)
    border_color: Color = Color(0, 255, 0, 255)
    border_width: float = 2


@dataclass()
class RectanglePrim(Primitive):
    name: ClassVar[str] = 'rectangle'
    fill_color: Color = Color(255, 0, 0, 255)
    border_color: Color = Color(0, 255, 0, 255)
    border_width: float = 2
    corner_radius: float = 4


class ChoiceItem(NamedTuple):
    value: Any
    text: str


class TextAlignment(Enum):
    LEFT = auto()
    CENTER = auto()
    RIGHT = auto()


FONT_FAMILY_CHOICES = [
    ChoiceItem(wx.FONTFAMILY_SWISS, 'sans-serif'),
    ChoiceItem(wx.FONTFAMILY_ROMAN, 'serif'),
    ChoiceItem(wx.FONTFAMILY_MODERN, 'monospace'),
    # ChoiceItem(wx.FONTFAMILY_DEFAULT, 'default'),
    # ChoiceItem(wx.FONTFAMILY_DECORATIVE, 'decorative'),
    # ChoiceItem(wx.FONTFAMILY_SCRIPT, 'script'),
]

FONT_STYLE_CHOICES = [
    ChoiceItem(wx.FONTSTYLE_NORMAL, 'normal'),
    ChoiceItem(wx.FONTSTYLE_ITALIC, 'italic'),
    # ChoiceItem(wx.FONTSTYLE_SLANT, 'slant'),
]

FONT_WEIGHT_CHOICES = [
    ChoiceItem(wx.FONTWEIGHT_MEDIUM, 'normal'),
    ChoiceItem(wx.FONTWEIGHT_BOLD, 'bold'),
    ChoiceItem(wx.FONTWEIGHT_LIGHT, 'light'),
]

TEXT_ALIGNMENT_CHOICES = [
    ChoiceItem(TextAlignment.LEFT, 'left'),
    ChoiceItem(TextAlignment.CENTER, 'center'),
    ChoiceItem(TextAlignment.RIGHT, 'right'),
]

class TextPosition(Enum):
    IN_NODE = auto()
    ABOVE = auto()
    BELOW = auto()
    NEXT_TO = auto()

TEXT_POSITION_CHOICES = [
  ChoiceItem(TextPosition.IN_NODE, "Text Inside"),
  ChoiceItem(TextPosition.ABOVE, "Above Node"),
  ChoiceItem(TextPosition.BELOW, "Below Node"),
  ChoiceItem(TextPosition.NEXT_TO, "Next to Node")
]


@dataclass()
class TextPrim(Primitive):
    name: ClassVar[str] = 'text'
    bg_color: Color = Color(255, 255, 0, 0)
    font_color: Color = Color(0, 0, 0, 255)
    font_size: int = 11
    font_family: int = wx.FONTFAMILY_SWISS
    font_style: int = wx.FONTSTYLE_NORMAL
    font_weight: int = wx.FONTWEIGHT_MEDIUM
    alignment: TextAlignment = TextAlignment.CENTER
    position: TextPosition = TextPosition.IN_NODE


def gen_polygon_pts(n, r=0.5, phase=0) -> Tuple[Vec2, ...]:
    """
    This function is used to define the vertices in 2D space of n-polygons. Each equilateral
    polygon is drawn inside a circle with specified radius.

    n: the number of sides of the polygon
    r: radius of the circle in which the polygon is drawn
    phase: the phase of the first point, in radians. If phase is 0, the first point is drawn at
           (r, 0) relative to the origin.
    """
    assert n >= 2

    origin = Vec2()
    inc = 2 * pi /n

    return tuple(origin + r * Vec2(cos(inc * i + phase), sin(inc * i + phase))
                 for i in range(n + 1))


@dataclass()
class PolygonPrim(Primitive):
    name: ClassVar[str] = 'polygon'
    points: Tuple[Vec2, ...]
    fill_color: Color = Color(255, 0, 0, 255)
    border_color: Color = Color(0, 255, 0, 255)
    border_width: float = 2
    radius: float = 0.5


@dataclass()
class HexagonPrim(PolygonPrim):
    name: ClassVar[str] = 'hexagon'
    points: Tuple[Vec2, ...] = gen_polygon_pts(6)


@dataclass()
class LinePrim(PolygonPrim):
    name: ClassVar[str] = 'line'
    # exclude the last point since we don't need the lines to be closed
    points: Tuple[Vec2, ...] = gen_polygon_pts(2)[:-1]


@dataclass()
class TrianglePrim(PolygonPrim):
    name: ClassVar[str] = 'triangle'
    points: Tuple[Vec2, ...] = gen_polygon_pts(3)

@dataclass()
class CompositeShape:
    items: List[Tuple[Primitive, Transform]]
    text_item: Tuple[TextPrim, Transform]
    name: str

    def __copy__(self):
        return CompositeShape(copy.deepcopy(self.items), copy.deepcopy(self.text_item), self.name)


class PrimitiveFactory:
    '''Factory that produces primitives.

    Why do we need this?
        1. Dynamic population of fields. When I need a primitive, I call produce(), which creates
           a primitive based on the latest data, e.g. theme.
        2. That's it.

    How do we achieve dynamically created fields? In the constructor of this class, we allow
    passing functions as values for fields. On product(), we call these functions to populate the
    fields. Of course, the user can still pass all-static fields, which would work as intended.
    '''
    def __init__(self, prim_class, **kwargs):
        fields = prim_class.__dataclass_fields__

        for key in kwargs.keys():
            if key not in fields:
                assert False  # TODO raise ValueError
        self.prim_class = prim_class
        self.abstract_kwargs = kwargs

    def produce(self):
        kwargs = dict()
        # construct actual arguments
        for key, value in self.abstract_kwargs.items():
            concrete_value = None
            if callable(value):
                concrete_value = value()
            else:
                concrete_value = value
            kwargs[key] = concrete_value
        return self.prim_class(**kwargs)


class CompositeShapeFactory:
    '''A factory for a composite shape, see PrimitiveFactory for more information.
    '''
    def __init__(self, item_factories: List[Tuple[PrimitiveFactory, Transform]],
                 text_factory: Tuple[PrimitiveFactory, Transform], name: str):
        self.item_factories = item_factories
        self.text_factory = text_factory
        self.name = name

    def produce(self):
        items = [(prim.produce(), tf) for prim, tf in self.item_factories]
        textitem = (self.text_factory[0].produce(), self.text_factory[1])
        return CompositeShape(items, textitem, self.name)


class RectData:
    position: Vec2
    size: Vec2


class Node(RectData):
    """Class that represents a Node for rendering purposes.

    Attributes:
        index: The index of the node. If this node has not yet been added to the NOM, this takes on
               the value of -1.
        id: The ID of the node.
        fill_color: The fill color of the node.
        border_color: The border color of the node.
        border_width: The border width of the node.
        position: Position of the size.
        size: Position of the size.
        net_index: The network index of the node.
    """
    id: str
    # fill_color: wx.Colour
    # border_color: wx.Colour
    # border_width: float
    concentration: float
    node_name: str
    node_SBO: str
    position: Vec2
    size: Vec2
    comp_idx: int
    index: int
    net_index: int
    floatingNode: bool
    lockNode: bool  # Prevent users from moving the node
    shape_index: int
    composite_shape: Optional[CompositeShape]
    # -1 if this is an original node, or if this is an alias node, then the index of the original copy
    original_index: int


    # force keyword-only arguments
    def __init__(self, id: str, net_index: int, *, pos: Vec2, size: Vec2, comp_idx: int = -1,
                 floatingNode: bool = True,
                 lockNode: bool = False,
                 shape_index: int = 0,
                 composite_shape: Optional[CompositeShape] = None,
                 index: int = -1,
                 original_index: int = -1,
                 concentration: float = 0.0, 
                 node_name: str = '',
                 node_SBO: str = ''):
        self.index = index
        self.net_index = net_index
        self.id = id
        self.position = pos
        self.size = size
        # self.fill_color = fill_color
        # self.border_color = border_color
        # self.border_width = border_width
        self.comp_idx = comp_idx
        self.floatingNode = floatingNode
        self.lockNode = lockNode
        self.shape_index = shape_index
        self.composite_shape = composite_shape
        self.original_index = original_index
        self.concentration = concentration
        self.node_name = node_name
        self.node_SBO = node_SBO

    def _get_prim_field(self, field):
        for prim, _ in self.composite_shape.items:
            if hasattr(prim, field):
                return getattr(prim, field)
        return None

    @property
    def fill_color(self):
        return self._get_prim_field('fill_color')

    @property
    def border_color(self):
        return self._get_prim_field('border_color')

    @property
    def border_width(self):
        return self._get_prim_field('border_width')

    @property
    def s_position(self):
        """The scaled position of the node obtained by multiplying the scale."""
        return self.position

    @property
    def s_size(self):
        """The scaled size of the node obtained by multiplying the scale."""
        return self.size

    @property
    def s_rect(self):
        """Return scaled position/size as Rect.

        Note that the fields of the returned rect is copied, so one cannot modify this node through
        the rect.
        """
        return Rect(copy.copy(self.s_position), copy.copy(self.s_size))

    @property
    def rect(self):
        """The same as s_rect, but the rectangle is unscaled.
        """
        return Rect(self.position, self.size)

    def __repr__(self):
        return 'Node(index={}, id="{}")'.format(self.index, self.id)

    def props_equal(self, other: 'Node'):
        return self.id == other.id and self.position == other.position and \
            self.size == other.size


def compute_centroid(rects: Sequence[Rect]) -> Vec2:
    """Compute the centroid position of a list of reactant and product nodes."""
    total = sum((r.center_point for r in rects), Vec2())
    return total / (len(rects))


BezJ = np.zeros((MAXSEGS + 1, 5))  #: Precomputed Bezier curve data
BezJPrime = np.zeros((MAXSEGS + 1, 5))  #: Precomputed bezier curve data
INITIALIZED = False  #: Flag for asserting that the above data is initialized
CURVE_SLACK = 5  #: Distance allowed on either side of a curve for testing click hit.


# inefficient implementation of comb(), but we only call this on startup, so that's fine
def comb(n, k):
    if k < n-k:
        return reduce(mul, range(n-k+1, n+1), 1) // factorial(k)
    else:
        return reduce(mul, range(k+1, n+1), 1) // factorial(n-k)


def init_bezier():
    """Initialize (precompute) the Bezier data."""
    global INITIALIZED

    if not INITIALIZED:
        for ti in range(MAXSEGS+1):
            t = ti/MAXSEGS
            for i in range(4):  # i = 0, 1, 2, 3
                BezJ[ti, i] = cast(int, comb(3, i)) * math.pow(t, i) * math.pow(1-t, 3-i)
            # At the moment hard-wired for n = 3
            tm = 1 - t
            BezJPrime[ti, 0] = -3*tm*tm
            BezJPrime[ti, 1] = 3*tm*tm - 6*t*tm
            BezJPrime[ti, 2] = 6*t*tm - 3*t*t
            BezJPrime[ti, 3] = 3*t*t

        INITIALIZED = True


class ModifierTipStyle(Enum):
    CIRCLE = 'circle'
    TEE = 'tee'


@dataclass
class Reaction:
    id: str
    net_index: int
    index: int
    center_pos: Optional[Vec2]
    fill_color: wx.Colour
    rate_law: str
    _sources: List[int]
    _targets: List[int]
    _thickness: float
    src_c_handle: 'HandleData'
    dest_c_handle: 'HandleData'
    handles: List['HandleData']
    bezierCurves: bool
    modifiers: Set[int]
    modifier_tip_style: ModifierTipStyle

    def __init__(self, id: str, net_index: int, *, sources: List[int], targets: List[int],
                 handle_positions: List[Vec2], fill_color: wx.Colour,
                 line_thickness: float, rate_law: str, center_pos: Optional[Vec2] = None,
                 bezierCurves: bool = True, modifiers: Set[int] = None,
                 modifier_tip_style: ModifierTipStyle = ModifierTipStyle.CIRCLE, index: int = -1):
        """Constructor for a reaction.

        Args:
            id: Reaction ID.
            sources: List of source (reactant) nodes.
            targets: List of target (product) nodes.
            handle: List of HandleData structs for reactant nodes and product nodes. This is in the
                    same order as sources and targets, i.e. [*sources, *targets].
            src_c_handle: HandleData struct for the source centroid handle.
            dest_c_handle: HandleData struct for the dest centroid handle.
            fill_color: Fill color of the curve.
            rate_law: The rate law string of the reaction; may not be valid.
            index: Reaction index.
            net_index: The network index of the reaction.
        """
        self.id = id
        self.net_index = net_index
        self.index = index
        self.center_pos = center_pos
        self.fill_color = fill_color
        self.rate_law = rate_law
        self._sources = sources
        self._targets = targets
        self._thickness = line_thickness
        self.bezierCurves = bezierCurves
        self.modifier_tip_style = modifier_tip_style
        if modifiers is None:
            modifiers = set()
        self.modifiers = modifiers

        assert len(handle_positions) == len(sources) + len(targets) + 1
        src_handle_pos = handle_positions[0]
        # Initialize to some arbitrary value. This should be controlled by ReactionBezier later.
        dest_handle_pos = Vec2()
        handle_pos = handle_positions[1:]

        self.src_c_handle = HandleData(src_handle_pos)
        self.dest_c_handle = HandleData(dest_handle_pos)

        self.handles = [HandleData(p) for p in handle_pos]

    @property
    def thickness(self) -> float:
        return self._thickness

    @property
    def sources(self) -> List[int]:
        return self._sources

    @property
    def targets(self) -> List[int]:
        return self._targets


def paint_handle(gc: wx.GraphicsContext, base: Vec2, handle: Vec2, hovering: bool):
    """Paint the handle as given by its base and tip positions, highlighting it if hovering."""
    c = get_theme('highlighted_handle_color') if hovering else get_theme('handle_color')
    brush = wx.Brush(c)
    pen = gc.CreatePen(wx.GraphicsPenInfo(c))

    gc.SetPen(pen)

    # Draw handle lines
    gc.StrokeLine(*base, *handle)

    # Draw handle circles
    gc.SetBrush(brush)
    gc.DrawEllipse(handle.x - HANDLE_RADIUS, handle.y - HANDLE_RADIUS,
                   2 * HANDLE_RADIUS, 2 * HANDLE_RADIUS)


class HandleData:
    """Struct for keeping handle data in-sync between the BezierHandle element and the Reaction.

    Attributes:
        tip: The position of the tip of the handle. May be modified by BezierHandle when user drags
             the handle element.
        base: The position of the base of the handle. May be modified by ReactionBezier, etc. as
              a response to movement of nodes, handles, etc. HACK this is only updated in the
              do_paint method of ReactionElement, and since the BezierHandles are drawn after
              the ReactionElements, the position of the base *happens* to be updated each time, but
              if it were drawn before, it would be one step behind.
    """
    tip: Vec2
    base: Optional[Vec2]

    def __init__(self, tip: Vec2, base: Vec2 = None):
        self.tip = tip
        self.base = base


class SpeciesBezier:
    """Class that keeps track of the Bezier curve associated with a reaction species.

    Note:
        When drawing the reaction curve, the builtin AddCurveToPoint is used. However, for the
        purpose of detecting a click on the curve, we still need to compute a rough approximation
        of the Bezier curve (e.g. 5 straight line segments). Hence why we need to compute the Bezier
        curve.

    Attributes:
        node: The associated node.
        node_intersection: The intersection point between the handle and the padded node, i.e. the
                           point after which the handle is not drawn, to create a gap.
        handle: Beizer handle for the side on the species.
        cnetroid_handle: Bezier handle for the centroid, shared among all reactants/products of
                         this reaction.
        is_source: Whether this species is considered a source or a dest node.
        arrow_adjusted_coords: Coordinate array for the arrow vertices.
        bezierCurves: True if we are drawing Bezier curves. Otherwise we are drawing simple
                      straight lines.
    """
    node_idx: int
    node_intersection: Optional[Vec2]
    handle: HandleData
    centroid_handle: HandleData
    is_source: bool
    bezier_points: List[Vec2]
    real_center: Vec2
    thickness: float
    _extended_handle: Vec2
    _collision_dirty: bool  #: Whether the Bezier curve needs to be recomputed.
    _paint_dirty: bool
    bezierCurves: bool

    def __init__(self, node_idx: int, node_rect: Rect, handle: HandleData, real_center: Vec2,
                 centroid_handle: HandleData, is_source: bool, thickness: float, bezierCurves: bool):
        assert INITIALIZED, 'Bezier matrices not initialized! Call init_bezier()'
        self.node_idx = node_idx
        self.node_rect = node_rect
        self.node_intersection = None
        self.handle = handle
        self.centroid_handle = centroid_handle
        self.is_source = is_source
        self.bezier_points = [Vec2() for _ in range(MAXSEGS + 1)]
        self._collision_dirty = True
        self._paint_dirty = True
        self.update_curve(real_center)
        self.arrow_adjusted_coords = list()
        self.thickness = thickness
        self.bezierCurves = bezierCurves

    def update_curve(self, real_center: Vec2):
        """Called after either the node, the centroid, or at least one of their handles changed.
        """
        self.real_center = real_center
        self._collision_dirty = True
        self._paint_dirty = True

    def get_bounding_rect(self) -> Rect:
        self._recompute(True)
        top = min(pt.y for pt in self.bezier_points)
        bottom = max(pt.y for pt in self.bezier_points)
        left = min(pt.x for pt in self.bezier_points)
        right = max(pt.x for pt in self.bezier_points)
        return Rect(Vec2(left, top), Vec2(right - left, bottom - top))

    def _recompute(self, for_collision: bool):
        """Recompute everything that could have changed.

        for_collision is an attempt at optimizing this function. If it is True, then the reaction
        curve segments will be recalculated (the segments are for detecting if the user clicked
        on them).
        """
        if self._collision_dirty or self._paint_dirty:
            # STEP 1, get intersection between handle and node outer padding
            node_center = self.node_rect.position + self.node_rect.size / 2

            # add a padding to node, so that the Bezier curve starts outside the node. The portion
            # of the handle within this is not drawn
            outer_rect = padded_rect(self.node_rect, NODE_EDGE_GAP_DISTANCE)

            self.node_intersection = None
            # extend handle to make sure that an intersection is found between the handle and the
            # node rectangle sides. We're essentially making the handle a ray.
            longer_side = max(outer_rect.size.x, outer_rect.size.y)
            long_dist = (longer_side + NODE_EDGE_GAP_DISTANCE) * 10

            other_point = self.handle.tip if self.bezierCurves else self.real_center
            handle_diff = other_point - node_center
            if handle_diff.norm_sq <= 1e-6:
                # if norm is too small, make it a hardcoded value to avoid zero vectors
                handle_diff = Vec2(0, 1)
            self._extended_handle = node_center + handle_diff.normalized(long_dist)
            handle_segment = (node_center, self._extended_handle)
            # check for intersection on the side
            self.node_intersection = segment_rect_intersection(handle_segment, outer_rect)

            assert self.node_intersection is not None

        if for_collision:
            # STEP 2, recompute Bezier curve
            if self._collision_dirty:
                for i in range(MAXSEGS+1):
                    tmp = Vec2()
                    for j, point in enumerate((self.node_intersection, self.handle.tip,
                                               self.centroid_handle.tip, self.real_center)):
                        tmp += point * float(BezJ[i, j])

                    # and scale back down again
                    self.bezier_points[i] = tmp

                self._collision_dirty = False
        else:
            # STEP 3, recompute arrow tip
            if self._paint_dirty:
                self._paint_dirty = False
                if not self.is_source:
                    self._recompute_arrow_tip(self.node_intersection,
                                              self.node_intersection - self._extended_handle)

    def arrow_tip_changed(self):
        self._paint_dirty = True

    def _recompute_arrow_tip(self, tip, slope):
        """Helper that recomputes the vertex coordinates of the arrow, given the tip pos and slope.
        """
        alpha = -math.atan2(slope.y, slope.x)
        cosine = math.cos(alpha)
        sine = math.sin(alpha)

        # Adjust the tip so that it moves forward slightly
        tip += TIP_DISPLACEMENT * Vec2(cosine, -sine)

        # Rotate the arrow into the correct orientation
        self.arrow_adjusted_coords = list()
        points = cstate.arrow_tip.points
        for i in range(4):
            coord = Vec2(points[i].x * cosine + points[i].y * sine,
                         -points[i].x * sine + points[i].y * cosine)
            self.arrow_adjusted_coords.append(coord)

        # Compute the distance of the tip of the arrow to the end point on the line
        # where the arrow should be placed. Then use this distance to translate the arrow
        offset = tip - self.arrow_adjusted_coords[3]
        # Translate the remaining coordinates of the arrow, note tip = Q
        for i in range(4):
            self.arrow_adjusted_coords[i] += offset

    def is_on_curve(self, pos: Vec2) -> bool:
        """Check if position is on curve; pos is scaled logical position."""
        self._recompute(for_collision=True)
        if self.bezierCurves:
            return any(pt_on_line(p1, p2, pos,
                                  CURVE_SLACK + self.thickness / 2)
                       for p1, p2 in pairwise(self.bezier_points))
        else:
            assert self.node_intersection is not None
            return pt_on_line(self.node_intersection,
                              self.real_center, pos,
                              CURVE_SLACK + self.thickness / 2)

    def do_paint(self, gc: wx.GraphicsContext, fill: wx.Colour, selected: bool):
        self._recompute(for_collision=False)
        rxn_color: wx.Colour
        # Draw bezier curve
        if selected:
            rxn_color = get_theme('selected_reaction_fill')
        else:
            rxn_color = fill

        pen = gc.CreatePen(wx.GraphicsPenInfo(rxn_color).Width(self.thickness))

        gc.SetPen(pen)
        # gc.StrokeLines([wx.Point2D(*(p * cstate.scale)) for p in self.bezier_points])
        path = gc.CreatePath()
        points = [p for p in (self.node_intersection,
                              self.handle.tip,
                              self.centroid_handle.tip,
                              self.real_center)]
        path.MoveToPoint(*points[0])
        if self.bezierCurves:
            path.AddCurveToPoint(*points[1], *points[2], *points[3])
        else:
            path.AddLineToPoint(*points[3])
        gc.StrokePath(path)

        if selected:
            assert self.node_intersection is not None
            self.handle.base = self.node_intersection

        # Draw arrow tip
        if not self.is_source:
            color = get_theme('handle_color') if selected else fill
            self.paint_arrow_tip(gc, color)

    def paint_arrow_tip(self, gc: wx.GraphicsContext, fill: wx.Colour):
        assert len(self.arrow_adjusted_coords) == 4, \
            "Arrow adjusted coords is not of length 4: {}".format(self.arrow_adjusted_coords)
        gc.SetPen(gc.CreatePen(wx.GraphicsPenInfo(fill)))
        gc.SetBrush(wx.Brush(fill))
        gc.DrawLines([wx.Point2D(*(coord))
                      for coord in self.arrow_adjusted_coords])


class ReactionBezier:
    """Class that keeps track of all Bezier curve data for a reaction.

    Attributes:
        TODO move this documentation to utils
        CENTER_RATIO: The ratio of source centroid handle length to the distance between the zeroth
                      node and the centroid (center handle is aligned with the zeroth node)
        DUPLICATE_RATIO: Valid for a Bezier whose node is both a product and a reactant. The ratio
                         of length of the product Bezier handle to the distance between the node
                         and the centroid.
        DUPLICATE_ROT: Rotation (radians) applied to the product Bezier handle, for nodes that are
                       both reactant and product. The handle is rotated so as to not align perfectly
                       with the source Bezier handle (otherwise the reactant and product curves
                       would completely overlap).
        src_beziers: List of SpeciesBezier instances for reactants.
        dest_beziers: List of SpeciesBezier instances for products.
        TODO move this to Reaction
        src_c_handle: Centroid bezier handle that controls the reactant curves.
        dest_c_handle: Centroid bezier handle that controls the product curves.
        handles: List of all the BezierHandle instances, stored for convenience.
    """
    src_beziers: List[SpeciesBezier]
    dest_beziers: List[SpeciesBezier]
    _bounding_rect: Rect

    def __init__(self, reaction: Reaction, reactants: List[Node], products: List[Node]):
        """Constructor.

        Args:
            reactants: List of reactant nodes.
            products: List of product nodes.
            handle_positions: The list of positions of the handles. 0th element is the position of
                              the reactant-side centroid handle, followed by the reactant handle
                              positions given in the same order as the reactants list, and
                              similarly followed by the product handle positins. Leave as None to
                              automatically initialize the handle positions (e.g. when a new
                              reaction is created).
        """
        assert len(reactants) == len(reaction.sources)
        assert len(products) == len(reaction.targets)
        self.reaction = reaction
        self.centroid = compute_centroid([n.rect for n in chain(reactants, products)])
        self.src_beziers = list()
        self.dest_beziers = list()

        reaction.dest_c_handle.tip = self.real_center * 2 - reaction.src_c_handle.tip
        reaction.src_c_handle.base = self.real_center
        reaction.dest_c_handle.base = self.real_center
        # create handles for species
        for index, (gi, node) in enumerate(gchain(reactants, products)):
            in_products = bool(gi)

            node_handle = reaction.handles[index]
            centroid_handle = self.reaction.dest_c_handle if in_products else self.reaction.src_c_handle
            sb = SpeciesBezier(node.index, node.rect, node_handle, self.real_center, centroid_handle,
                               not in_products, self.reaction.thickness, self.reaction.bezierCurves)
            to_append = self.dest_beziers if in_products else self.src_beziers
            to_append.append(sb)

    @property
    def real_center(self) -> Vec2:
        return self.reaction.center_pos if self.reaction.center_pos else self.centroid

    def make_handle_moved_func(self, sb: SpeciesBezier):
        """Manufacture a callback function (on_moved) for the given SpeciesBezier."""
        return lambda _: sb.update_curve(self.real_center)

    def is_mouse_on(self, pos: Vec2) -> bool:
        """Return whether mouse is on the Bezier curve (not including the handles).

        pos is the logical position of the mouse (and not multiplied by any scale).
        """
        return any(bz.is_on_curve(pos) for bz in chain(self.src_beziers, self.dest_beziers))

    def src_handle_moved(self):
        """Special callback for when the source centroid handle is moved."""
        self.reaction.dest_c_handle.tip = 2 * self.real_center - self.reaction.src_c_handle.tip
        for bz in chain(self.src_beziers, self.dest_beziers):
            bz.update_curve(self.real_center)

    def dest_handle_moved(self):
        """Special callback for when the dest centroid handle is moved."""
        self.reaction.src_c_handle.tip = 2 * self.real_center - self.reaction.dest_c_handle.tip
        for bz in chain(self.src_beziers, self.dest_beziers):
            bz.update_curve(self.real_center)

    def center_moved(self, offset: Vec2):
        self.reaction.src_c_handle.base = self.real_center
        self.reaction.dest_c_handle.base = self.real_center
        self.reaction.src_c_handle.tip += offset
        self.src_handle_moved()

    def nodes_moved(self, rects: List[Rect]):
        # TODO set center pos, but not controller. Need to update controller later.
        self.centroid = compute_centroid(rects)
        self.reaction.src_c_handle.base = self.real_center
        self.reaction.dest_c_handle.base = self.real_center
        for i, sb in enumerate(chain(self.src_beziers, self.dest_beziers)):
            sb.node_rect = rects[i]
        self.src_handle_moved()

    def do_paint(self, gc: wx.GraphicsContext, fill: wx.Colour, selected: bool):
        for bz in chain(self.src_beziers, self.dest_beziers):
            bz.do_paint(gc, fill, selected)

    def get_bounding_rect(self):
        return get_bounding_rect(
            list(sb.get_bounding_rect() for sb in chain(self.src_beziers, self.dest_beziers)))


@dataclass
class Compartment(RectData):
    id: str
    net_index: int
    nodes: List[int]
    volume: float  #: Size (i.e. length/area/volume/...) of the container
    position: Vec2
    size: Vec2  #: Size for drawing on the GUI
    fill: wx.Colour
    border: wx.Colour
    border_width: float
    index: int = -1
    #dimensions: int

    @property
    def rect(self):
        """The same as s_rect, but the rectangle is unscaled.
        """
        return Rect(self.position, self.size)



"""
Display a random network.

Version 1.0.1: Author: Jin Xu, Herbert M. Sauro (2020)

"""


import wx
from rkviewer.plugin.classes import PluginMetadata, WindowedPlugin, PluginCategory
from rkviewer.plugin import api
from rkviewer.plugin.api import Node, Vec2, Reaction
import math
import random as _random
import numpy as _np
import copy as _copy
from dataclasses import dataclass
import decimal



@dataclass
class DefaultValues:
    maxSpecies = 15
    maxReactions = 20
    probUniUniValue = 1
    probUniBiValue = 0
    probBiUniValue = 0
    probBiBiValue = 0
    randomSeed = 130


class RandomNetwork(WindowedPlugin):
    metadata = PluginMetadata(
        name='RandomNetwork',
        author='Jin Xu, Herbert M. Sauro',
        version='1.0.1',
        short_desc='Random network.',
        long_desc='Display a random network with certain number of species and reactions as input.',
        category=PluginCategory.MODELS
    )
    def __init__(self):
        """
        Initialize the RandomNetwork.

        Args:
            self

        """
        super().__init__()
        

    def create_window(self, dialog):
        """
        Create a window with several inputs and buttons.
        Args:
            self
            dialog
        """

        window = wx.Panel(dialog, pos=(5,100), size=(300, 320))

        numSpecs = wx.StaticText(window, -1, 'Number of Species:', (20,20))
        self.numSpecsText = wx.TextCtrl(window, -1, str(DefaultValues.maxSpecies), (160, 20), size=(100, -1))
        self.numSpecsText.SetInsertionPoint(0)
        self.numSpecsText.Bind(wx.EVT_TEXT, self.OnText_numSpecs)
        self.numSpecsValue = int(self.numSpecsText.GetValue())
       
        numRxns = wx.StaticText(window, -1, 'Number of Reactions:', (20,50))
        self.numRxnsText = wx.TextCtrl(window, -1, str(DefaultValues.maxReactions), (160, 50), size=(100, -1))
        self.numRxnsText.SetInsertionPoint(0)
        self.numRxnsText.Bind(wx.EVT_TEXT, self.OnText_numRxns)
        self.numRxnsValue = int(self.numRxnsText.GetValue())  
 
        probUniUni = wx.StaticText(window, -1, 'Probability of UniUni:', (20,90))      
        self.probUniUniText = wx.TextCtrl(window, -1, str(DefaultValues.probUniUniValue), (160, 90), size=(100, -1))
        self.probUniUniText.SetInsertionPoint(0)
        self.probUniUniText.Bind(wx.EVT_TEXT, self.OnText_UniUni)
        self.probUniUniValue = float(self.probUniUniText.GetValue())        
 
        probBiUni = wx.StaticText(window, -1, 'Probability of BiUni:', (20,120))
        self.probBiUniText = wx.TextCtrl(window, -1, str(DefaultValues.probBiUniValue), (160, 120), size=(100, -1))
        self.probBiUniText.SetInsertionPoint(0)
        self.probBiUniText.Bind(wx.EVT_TEXT, self.OnText_BiUni)
        self.probBiUniValue = float(self.probBiUniText.GetValue())

        probUniBi = wx.StaticText(window, -1, 'Probability of UniBi:', (20,150))
        self.probUniBiText = wx.TextCtrl(window, -1, str(DefaultValues.probUniBiValue), (160, 150), size=(100, -1))
        self.probUniBiText.SetInsertionPoint(0)
        self.probUniBiText.Bind(wx.EVT_TEXT, self.OnText_UniBi)
        self.probUniBiValue = float(self.probUniBiText.GetValue())

        probBiBi = wx.StaticText(window, -1, 'Probability of BiBi:', (20,180))
        self.probBiBiText = wx.TextCtrl(window, -1, str(DefaultValues.probBiBiValue), (160, 180), size=(100, -1))
        self.probBiBiText.SetInsertionPoint(0)
        self.probBiBiText.Bind(wx.EVT_TEXT, self.OnText_BiBi)
        self.probBiBiValue = float(self.probBiBiText.GetValue())

        randomSeed = wx.StaticText(window, -1, 'Random seed:', (20,210))
        randomSeed = wx.StaticText(window, -1, '0 means no seed setup', (20,230))
        self.randomSeedText = wx.TextCtrl(window, -1, str(DefaultValues.randomSeed), (160, 210), size=(100, -1))
        self.randomSeedText.SetInsertionPoint(0)
        self.randomSeedText.Bind(wx.EVT_TEXT, self.OnText_randomSeed)
        self.randomSeedValue = float(self.randomSeedText.GetValue())
 
        apply_btn = wx.Button(window, -1, 'Apply', (160, 250))
        apply_btn.Bind(wx.EVT_BUTTON, self.Apply)

        return window


    def OnText_numSpecs(self, evt):
        update = evt.GetString()
        if update != '':
            try:
                self.numSpecsValue = int(self.numSpecsText.GetValue())
            except:
                wx.MessageBox("Please enter an integer for the number of species.", "Message", wx.OK | wx.ICON_INFORMATION)


    def OnText_numRxns(self, evt):
        update = evt.GetString()
        if update != '':
            try:
                self.numRxnsValue = int(self.numRxnsText.GetValue())
            except:
                wx.MessageBox("Please enter an integer for the number of reactions.", "Message", wx.OK | wx.ICON_INFORMATION)

 
    def OnText_UniUni(self, evt):
        update = evt.GetString()
        if update != '':
            try:
                self.probUniUniValue = float(self.probUniUniText.GetValue())
            except:
                wx.MessageBox("Please enter a floating point number for the probability of UniUni.", "Message", wx.OK | wx.ICON_INFORMATION)


    def OnText_BiUni(self, evt):
        update = evt.GetString()
        if update != '':
            #DefaultValues.probBiUniValue = float(self.probBiUniText.GetValue())
            try:
                self.probBiUniValue = float(self.probBiUniText.GetValue())
            except:
                wx.MessageBox("Please enter a floating point number for the probability of BiUni.", "Message", wx.OK | wx.ICON_INFORMATION)

    def OnText_UniBi(self, evt):
        update = evt.GetString()
        if update != '':
            try:
                self.probUniBiValue = float(self.probUniBiText.GetValue())
            except:
                wx.MessageBox("Please enter a floating point number for the probability of UniBi.", "Message", wx.OK | wx.ICON_INFORMATION)


    def OnText_BiBi(self, evt):
        update = evt.GetString()
        if update != '':
            try:
                self.probBiBiValue = float(self.probBiBiText.GetValue())
            except:
                wx.MessageBox("Please enter a floating point number for the probability of BiBi.", "Message", wx.OK | wx.ICON_INFORMATION)

    def OnText_randomSeed(self, evt):
        update = evt.GetString()
        if update != '':
            try:
                self.randomSeedValue = float(self.randomSeedText.GetValue())
            except:
                wx.MessageBox("Please enter a valid random seed other than zero.", "Message", wx.OK | wx.ICON_INFORMATION)


    def Apply(self, evt):
        """
        Handler for the "apply" button. apply the random network.
        """
        if self.randomSeedValue != 0:
            _random.seed(self.randomSeedValue)

        class _TReactionType:
            UNIUNI = 0
            BIUNI = 1
            UNIBI = 2
            BIBI = 3

        def _pickReactionType():

            rt = _random.random()
            if rt < self.probUniUniValue:
                return _TReactionType.UNIUNI
            elif rt < self.probUniUniValue + self.probBiUniValue:
                return _TReactionType.BIUNI
            elif rt < self.probUniUniValue + self.probBiUniValue + self.probUniBiValue:
                return _TReactionType.UNIBI
            else:
                return _TReactionType.BIBI


        # Generates a reaction network in the form of a reaction list
        # reactionList = [nSpecies, reaction, reaction, ....]
        # reaction = [reactionType, [list of reactants], [list of products], rateConsta>
        # Disallowed reactions:
        # S1 -> S1
        # S1 + S2 -> S2  # Can't have the same reactant and product
        # S1 + S1 -> S1
        def _generateReactionList (nSpecies, nReactions):

            reactionList = []
            for r in range(nReactions):
       
                rateConstant = _random.random()
                rt = _pickReactionType()
                if rt ==  _TReactionType.UNIUNI:
                    # UniUni
                    reactant = _random.randint (0, nSpecies-1)
                    product = _random.randint (0, nSpecies-1)
                    # Disallow S1 -> S1 type of reaction
                    while product == reactant:
                        product = _random.randint (0, nSpecies-1)
                    reactionList.append ([rt, [reactant], [product], rateConstant])
               
                if rt ==  _TReactionType.BIUNI:
                    # BiUni
                    # Pick two reactants
                    reactant1 = _random.randint (0, nSpecies-1)
                    reactant2 = _random.randint (0, nSpecies-1)
               
                    # pick a product but only products that don't include the reactants
                    species = range (nSpecies)
                    # Remove reactant1 and 2 from the species list
                    species = _np.delete (species, [reactant1, reactant2], axis=0)
                    # Then pick a product from the reactants that are left
                    product = species[_random.randint (0, len (species)-1)]
               
                    reactionList.append ([rt, [reactant1, reactant2], [product], rateConstant])

                if rt ==  _TReactionType.UNIBI:
                    # UniBi
                    reactant1 = _random.randint (0, nSpecies-1)
           
               
                    # pick a product but only products that don't include the reactant
                    species = range (nSpecies)
                    # Remove reactant1 from the species list
                    species = _np.delete (species, [reactant1], axis=0)
                    # Then pick a product from the reactants that are left
                    product1 = species[_random.randint (0, len (species)-1)]
                    product2 = species[_random.randint (0, len (species)-1)]
   
                    reactionList.append ([rt, [reactant1], [product1, product2], rateConstant])

                if rt ==  _TReactionType.BIBI:
                    # BiBi
                    reactant1 = _random.randint (0, nSpecies-1)
                    reactant2= _random.randint (0, nSpecies-1)
               
                    # pick a product but only products that don't include the reactant
                    species = range (nSpecies)
                    # Remove reactant1 and 2 from the species list
                    species = _np.delete (species, [reactant1, reactant2], axis=0)
                    # Then pick a product from the reactants that are left
                    product1 = species[_random.randint (0, len (species)-1)]
                    product2 = species[_random.randint (0, len (species)-1)]
               
                    element = [rt, [reactant1, reactant2], [product1, product2], rateConstant]
                    reactionList.append (element)            

            reactionList.insert (0, nSpecies)
            return reactionList



        # Includes boundary and floating species
        # Returns a list:
        # [New Stoichiometry matrix, list of floatingIds, list of boundaryIds]
        def _getFullStoichiometryMatrix (reactionList):
     
            nSpecies = reactionList[0]
            reactionListCopy = _copy.deepcopy (reactionList)
            reactionListCopy.pop (0)
            st = _np.zeros ((nSpecies, len(reactionListCopy)))
   
            for index, r in enumerate (reactionListCopy):
                if r[0] ==  _TReactionType.UNIUNI:
                    # UniUni
                    reactant = reactionListCopy[index][1][0]
                    st[reactant, index] = -1
                    product = reactionListCopy[index][2][0]
                    st[product, index] = 1
     
                if r[0] ==  _TReactionType.BIUNI:
                    # BiUni
                    reactant1 = reactionListCopy[index][1][0]
                    st[reactant1, index] = -1
                    reactant2 = reactionListCopy[index][1][1]
                    st[reactant2, index] = -1
                    product = reactionListCopy[index][2][0]
                    st[product, index] = 1

                if r[0] ==  _TReactionType.UNIBI:
                    # UniBi
                    reactant1 = reactionListCopy[index][1][0]
                    st[reactant1, index] = -1
                    product1 = reactionListCopy[index][2][0]
                    st[product1, index] = 1
                    product2 = reactionListCopy[index][2][1]
                    st[product2, index] = 1
 
                if r[0] ==  _TReactionType.BIBI:
                    # BiBi
                    reactant1 = reactionListCopy[index][1][0]
                    st[reactant1, index] = -1
                    reactant2 = reactionListCopy[index][1][1]
                    st[reactant2, index] = -1
                    product1 = reactionListCopy[index][2][0]
                    st[product1, index] = 1
                    product2 = reactionListCopy[index][2][1]
                    st[product2, index] = 1

            return st

        def _getRateLaw (floatingIds, boundaryIds, reactionList, isReversible):
   
            nSpecies = reactionList[0]
            # Remove the first element which is the nSpecies
            reactionListCopy = _copy.deepcopy (reactionList)
            reactionListCopy.pop (0)

            antStr_tot = []

            for index, r in enumerate (reactionListCopy):
                antStr= ''
                antStr = antStr + 'J' + str (index) + ': '
                if r[0] == _TReactionType.UNIUNI:
                    # UniUni
                    antStr = antStr + '(k' + str (index) + '*S' + str (reactionListCopy[index][1][0])
                    if isReversible:
                        antStr = antStr + ' - k' + str (index) + 'r' + '*S' + str (reactionListCopy[index][2][0])
                    antStr = antStr + ')'
                if r[0] == _TReactionType.BIUNI:
                    # BiUni
                    antStr = antStr + '(k' + str (index) + '*S' + str (reactionListCopy[index][1][0]) + '*S' + str (reactionListCopy[index][1][1])
                    if isReversible:
                        antStr = antStr + ' - k' + str (index) + 'r' + '*S' + str (reactionListCopy[index][2][0])
                    antStr = antStr + ')'
                if r[0] == _TReactionType.UNIBI:
                    # UniBi
                    antStr = antStr + '(k' + str (index) + '*S' + str (reactionListCopy[index][1][0])
                    if isReversible:
                        antStr = antStr + ' - k' + str (index) + 'r' + '*S' + str (reactionListCopy[index][2][0]) + '*S' + str (reactionListCopy[index][2][1])
                    antStr = antStr + ')'
                if r[0] == _TReactionType.BIBI:
                    # BiBi
                    antStr = antStr + '(k' + str (index) + '*S' + str (reactionListCopy[index][1][0]) + '*S' + str (reactionListCopy[index][1][1])
                    if isReversible:
                        antStr = antStr + ' - k' + str (index) + 'r' + '*S' + str (reactionListCopy[index][2][0]) + '*S' + str (reactionListCopy[index][2][1])
                    antStr = antStr + ')'
 
                antStr_tot.append(antStr)

            return antStr_tot      

        decimal.getcontext().prec = 6
        test_prob = decimal.Decimal(self.probUniUniValue) \
                  + decimal.Decimal(self.probBiUniValue) \
                  + decimal.Decimal(self.probUniBiValue) \
                  + decimal.Decimal(self.probBiBiValue)

        if test_prob != 1:
            wx.MessageBox("The sum of probabilities should be one!", "Message", wx.OK | wx.ICON_INFORMATION)

        else:
            net_index = 0
            api.clear_network(net_index)

            rl = _generateReactionList (self.numSpecsValue, self.numRxnsValue)
            st = _getFullStoichiometryMatrix (rl)
            antStr = _getRateLaw (st[1], st[2], rl, isReversible=True)
            numNodes = st.shape[0]
            numRxns = st.shape[1]

            nodeIdx = []
            for i in range (numNodes):
                nodeIdx.append (api.add_node(net_index, 'node_{}'.format(i), size=Vec2(60,40), 
                fill_color=api.Color(255, 204, 153),border_color=api.Color(255, 108, 9),
                position=Vec2(40 + math.trunc (_random.random()*800), 40 + math.trunc (_random.random()*800))))
        
            for i in range (numRxns):
                src = []
                dest = []
            
                for j in range(numNodes):
                    if (st.item(j,i) == -1):
                        src.append(nodeIdx[j])  
                    if (st.item(j,i) == 1):
                        dest.append(nodeIdx[j])
                r_idx = api.add_reaction(net_index, 'reaction_{}'.format(i), src, dest, fill_color=api.Color(91, 176, 253))
            
            # Need to remove orphan nodes
            for i in range (numNodes):
                if _np.array_equal(st[i,:], _np.zeros(numRxns)):
                    api.delete_node(net_index, nodeIdx[i])
        




"""Utility functions"""
# pylint: disable=maybe-no-member
import wx
from itertools import tee
import os
import platform
import subprocess
import sys
from typing import Any, Callable, Dict, Iterable, List, Tuple, Type, TypeVar
from dataclasses import is_dataclass
from pathlib import Path


def convert_position(fn):
    """Decorator that converts the event position to one that is relative to the receiver."""

    def ret(self, evt):
        if self is not evt.EventObject:
            client_pos = evt.GetPosition()  # get raw position
            screen_pos = evt.EventObject.ClientToScreen(client_pos)  # convert to screen position
            relative_pos = self.ScreenToClient(screen_pos)  # convert to receiver position
            # call function
            copy = evt.Clone()
            copy.SetPosition(relative_pos)
            copy.foreign = True
            fn(self, copy)
            evt.Skip()
        else:
            copy = evt
            copy.foreign = False
            fn(self, copy)

    return ret


def no_rzeros(num: float, precision: int) -> str:
    """Returns string of the num with the given precision, but with trailing zeros removed."""
    assert precision > 0
    fmt = '{:.' + str(precision) + 'f}'
    return fmt.format(num).rstrip('0').rstrip('.')


def on_msw() -> bool:
    """Returns whether we are running on Windows."""
    return os.name == 'nt'


def get_local_path(relative_path):
    """Get path relative to the executable, or the working dir if not bundled."""
    return os.path.abspath(relative_path)


# def get_bundled_path(relative_path):
#     """Given a path relative to the application bundle, return the absolute path.
    
#     Specifically for files bundled with the application, e.g. resources.
#     """
#     if hasattr(sys, '_MEIPASS'):
#         # PyInstaller creates a temp folder and stores path in _MEIPASS
#         base_path = getattr(sys, '_MEIPASS')
#     else:
#         base_path = os.path.abspath(".")

#     return os.path.join(base_path, relative_path)


def resource_path(relative_path):
    """Get absolute path to resource, works for dev and PyInstaller."""

    source_path = Path(__file__).resolve()
    source_dir = source_path.parent
    return source_dir.joinpath('resources').joinpath(relative_path).as_posix()


def start_file(abs_path: str):
    # Tell OS to open the file for editing. From https://stackoverflow.com/a/435669/9171534
    if platform.system() == 'Darwin':       # macOS
        subprocess.call(('open', abs_path))
    elif platform.system() == 'Windows':    # Windows
        os.startfile(abs_path)
    else:                                   # linux variants
        subprocess.call(('xdg-open', abs.path))


def pairwise(iterable: Iterable) -> Iterable:
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def rgba_to_wx_colour(rgb: int, alpha: float) -> wx.Colour:
    """Given RGBA color, return wx.Colour.

    Args:
        rgb: RGB color in hex format.
        alpha: The opacity of the color, ranging from 0.0 to 1.0.
    """
    b = rgb & 0xff
    g = (rgb >> 8) & 0xff
    r = (rgb >> 16) & 0xff
    return wx.Colour(r, g, b, int(alpha * 255))


def opacity_mul(color: wx.Colour, fraction: float) -> wx.Colour:
    return wx.Colour(color.Red(), color.Green(), color.Blue(), int(color.Alpha() * fraction))


def change_opacity(color: wx.Colour, new_op: int):
    return wx.Colour(color.Red(), color.Green(), color.Blue(), new_op)


class ButtonGroup:
    """Class for keeping track of a group of buttons, where exactly one of them can be selected.

    Attributes:
        Callback: The callback function type called with ID of the selected button as argument.
    """
    Callback = Callable[[str], None]

    def __init__(self, callback: Callback):
        """Construct a ButtonGroup.

        Args:
            callback: The callback function called when a new button is selected.
        """
        self.callback = callback
        self.buttons = []
        self.selected = None  # should be tuple (button, group_id)

    def AddButton(self, button: wx.ToggleButton, identifier: Any):
        """Add a button with the given identifier.
        
        When this button is clicked, callback is called with the identifier.
        """
        # right now there is no type info for wxPython, so this is necessary
        assert isinstance(button, wx.ToggleButton)

        self.buttons.append(button)
        button.Bind(wx.EVT_TOGGLEBUTTON, self._MakeToggleFn(button, identifier))

        # First added button; make it selected
        if self.selected is None:
            self.selected = (button, identifier)
            button.SetValue(True)
            self.callback(identifier)

    def _MakeToggleFn(self, button: wx.ToggleButton, group_id: Any):
        """Create the function to be called by a specific button in the group when it is clicked.
        """
        # right now there is no type info for wxPython, so this is necessary
        assert isinstance(button, wx.ToggleButton)

        def ret(evt):
            assert self.selected is not None, "There must be at least one button in ButtonGroup!"

            if evt.IsChecked():
                button.SetValue(True)
                selected_btn, selected_id = self.selected
                if selected_id != group_id:
                    selected_btn.SetValue(False)
                    self.selected = (button, group_id)
                    self.callback(group_id)
            else:
                # don't allow de-select
                button.SetValue(True)
        return ret


def int_round(n: float) -> int:
    return int(round(round(n, 2)))


def even_round(n: float) -> int:
    """Round to the nearest even integer"""
    return int(round(n / 2)) * 2


T = TypeVar('T')
def gchain(*iterables: Iterable[T]) -> Iterable[Tuple[int, T]]:
    # chain('ABC', 'DEF') --> (0,A) (0,B) (0,C) (1,D) (1,E) (1,F)
    i = 0
    for it in iterables:
        for element in it:
            yield (i, element)
        i += 1


# Force dataclass to use kwargs, from https://gist.github.com/mikeholler/4be180627d3f8fceb55704b729464adb#file-dataclass_utils-py
_T = TypeVar("_T")
_Self = TypeVar("_Self")
_VarArgs = List[Any]
_KWArgs = Dict[str, Any]


def _kwarg_only_init_wrapper(
        self: _Self,
        init: Callable[..., None],
        *args: _VarArgs,
        **kwargs: _KWArgs
) -> None:
    if len(args) > 0:
        raise TypeError(
            f"{type(self).__name__}.__init__(self, ...) only allows keyword arguments. Found the "
            f"following positional arguments: {args}"
        )
    init(self, **kwargs)


def require_kwargs_on_init(cls: Type[_T]) -> Type[_T]:
    """
    Force a dataclass's init function to only work if called with keyword arguments.
    If parameters are not positional-only, a TypeError is thrown with a helpful message.
    This function may only be used on dataclasses.
    This works by wrapping the __init__ function and dynamically replacing it. Therefore,
    stacktraces for calls to the new __init__ might look a bit strange. Fear not though,
    all is well.
    Note: although this may be used as a decorator, this is not advised as IDEs will no longer
    suggest parameters in the constructor. Instead, this is the recommended usage::
        from dataclasses import dataclass
        @dataclass
        class Foo:
            bar: str
        require_kwargs_on_init(Foo)
    """
    if cls is None:
        raise TypeError("Cannot call with cls=None")
    if not is_dataclass(cls):
        raise TypeError(
            f"This decorator only works on dataclasses. {cls.__name__} is not a dataclass."
        )

    original_init = cls.__init__

    def new_init(self: _Self, *args: _VarArgs, **kwargs: _KWArgs) -> None:
        _kwarg_only_init_wrapper(self, original_init, *args, **kwargs)

    # noinspection PyTypeHints
    cls.__init__ = new_init  # type: ignore

    return cls
